"""
Data, figures/tables IO handy utilities.
"""

import os as _os
import pandas as _pd
import tempfile as _tempfile
import subprocess as _subprocess
import zipfile as _zipfile
import logging as _logging
import shutil as _shutil

from io import StringIO as _StringIO
from io import BytesIO as _BytesIO

# from collections import OrderedDict as _OrderedDict
from pathlib import Path as _Path
from pylbmisc.dm import _default_dtype_backend
from pylbmisc.dm import fix_varnames as _fix_varnames
from pylbmisc.dm import is_bool as _is_bool
from pylbmisc.dm import is_categorical as _is_categorical
from pylbmisc.dm import is_date as _is_date
from pylbmisc.dm import is_datetime as _is_datetime
from pylbmisc.dm import is_integer as _is_integer
from pylbmisc.dm import is_numeric as _is_numeric
from pylbmisc.dm import is_string as _is_string
from pylbmisc.dm import is_all_missing as _is_all_missing
from pylbmisc.dm import to_categorical as _to_categorical
from pylbmisc.dm import to_date as _to_date
from typing import Sequence as _Sequence
from pylbmisc.r import match_arg as _match_arg

# ------------------------------------
# Figure and images stuff
# ------------------------------------
def export_figure(fig,
                  fdir: str = "outputs",
                  fname: str = "",
                  fext: str | list | set = ("eps", "png", "pdf"),
                  latex_include=True,
                  label: str = "",
                  caption: str = "",
                  scale: float = 1
                  ) -> None:
    """Dump a figure (save to file, include in LaTeX)

    Save to png, eps and pdf and include in Latex an image;
    intended to be used inside pythontex pycode.

    Parameters
    ----------
    fig: matplotlib.figure.Figure
        the fig
    fdir: str
        directory dove salvare, se mancante si usa /tmp
    fname: str
        basename del file da salvare, se mancante si prova a riusare
        label (se specificata) oppure a creare un file temporaneo
    fext: str, list or set
        list or set with format/extensions to be used for exporting
    latex_include: bool
        print command for latex includeing
    label: str
        posta dopo "fig:" in LaTeX
    caption: str
        caption LaTeX, se mancante viene riadattata label
    scale: float
        scale di includegraphics di LaTeX
    """
    # fdir not existing, using /tmp
    if not _os.path.isdir(fdir):
        fdir = "/tmp"

    if isinstance(fext, str):
        fext = [fext]

    # default filename to be set as label, if available or a temporary one
    if fname == "":
        if label != "":
            fname = label
        else:
            tmp = _tempfile.mkstemp(dir=fdir)
            fname = _os.path.basename(tmp[1])

    # produce the file paths
    base_path = _os.path.join(fdir, fname)
    eps_path = base_path + ".eps"
    png_path = base_path + ".png"
    pdf_path = base_path + ".pdf"

    # save figures to hard drive
    if "eps" in fext:
        fig.savefig(eps_path, transparent=True)
    if "png" in fext:
        fig.savefig(png_path, transparent=True)  # , dpi = 600)
    if "pdf" in fext:
        fig.savefig(pdf_path, transparent=True)

    # latex stuff
    if latex_include:
        latex_label = "fig:" + label
        caption = (
            label.capitalize().replace("_", " ") if caption == "" else caption
        )
        latex = (
            "\\begin{figure} \\centering "
            "\\includegraphics[scale=%(scale).2f]{%(base_path)s}"
            " \\caption{%(caption)s} \\label{%(label)s} \\end{figure}"
        )
        subs = {
            "scale": scale,
            "base_path": base_path,
            "caption": caption,
            "label": latex_label,
        }
        print(latex % subs)


# ------------------------------------
# Dataset Import/export routines
# ------------------------------------

def _make_rec_dict(f, t):
    rec_df = (
        _pd.DataFrame({"from": f, "to": t})
        .drop_duplicates()
        .dropna()
        .sort_values(by="from")
    )
    ft = dict()
    for row in rec_df.itertuples():
        ft[row[1]] = row[2]
    return ft


def _import_csv(path, csv_kwargs):
    """
    Import csv handling encryption for .csv.gpg files
    """
    fext = "".join(path.suffixes)
    if fext == ".csv.gpg":
        decrypt_cmd = f"gpg --decrypt {path}"
        decrypted_content = _subprocess.run(
            decrypt_cmd.split(" "),
            capture_output=True,
            text=True).stdout
        data_input = _StringIO(decrypted_content)
    else:
        data_input = path
    data = _pd.read_csv(data_input, **csv_kwargs)
    return data


def _import_excel(path, excel_kwargs):
    """
    Import csv handling encryption for .xls.gpg or .xlsx.gpg files
    """
    fext = "".join(path.suffixes)
    fname = path.name.replace(fext, "")
    if fext in {".xls.gpg", ".xlsx.gpg"}:
        decrypt_cmd = f"gpg --decrypt {path}"
        decrypted_content = _subprocess.run(
            decrypt_cmd.split(" "),
            capture_output=True).stdout
        data_input = _BytesIO(decrypted_content)
    else:
        data_input = path
    # import all the sheets as a dict of DataFrame
    sheets = _pd.read_excel(data_input, None, **excel_kwargs)
    # add xlsx name to sheets ones
    sheets = {f"{fname}_{k}": v for k, v in sheets.items()}
    return sheets


def import_redcap(
        data_fpath: str | _Path = "data/DATA.csv.gpg",
        labels_fpath: str | _Path = "data/LABELS.csv.gpg",
        csv_kwargs: dict = {"dtype_backend": _default_dtype_backend},
        verbose=False
) -> _pd.DataFrame:
    """
    Import dataset exported from redcap adding labels

    The idea is adding names as data['var'].name = labels['var'].column

    Parameters
    ----------
    data_path: str | Path
        csv (or csv.gpg) of raw data export
    labels_path: str | Path
        csv (or csv.gpg) of labelled data export

    Examples
    --------
    >>> df = import_redcap("data/DATA.csv.gpg", "data/LABELS.csv.gpg")
    """
    data = _import_csv(_Path(data_fpath), csv_kwargs)
    labels = _import_csv(_Path(labels_fpath), csv_kwargs)

    varnames = data.columns.to_list()
    comments = labels.columns.to_list()

    if (data.shape[1] != labels.shape[1]):
        msg = "I dataframe passati debbono avere lo stesso numero di colonne"
        raise Exception(msg)

    final_df = data.copy()
    description_dict = dict()

    # process column by column
    for ncol in range(data.shape[1]):
        varname = varnames[ncol]
        comment = comments[ncol]
        description_dict[varname] = comment
        if verbose:
            print(f"Doing {varname}")

        # do recoding if needed
        raw = data.iloc[:, ncol]    # raw variable (before recoding)
        rec = labels.iloc[:, ncol]  # recoded variable (after recoding)

        # factor: raw stringa/intero/numerico e ricodificato a stringa
        probable_factor = ((_is_integer(raw) or _is_numeric(raw))
                           and
                           _is_string(rec))
        probable_date = (_is_string(raw) and
                         raw.str.fullmatch(r"\d{4}-\d{2}-\d{2}").all())

        if probable_date:
            final_df[varname] = _to_date(raw)
        elif probable_factor:
            rec_dict = _make_rec_dict(raw, rec)
            final_df[varname] = _to_categorical(
                raw,
                levels=list(rec_dict.keys()),
                labels=list(rec_dict.values()))
        else:
            pass  # do nothing but in the previous cases

    return final_df, description_dict


def import_data(fpaths: str | _Path | _Sequence[str | _Path],
                csv_kwargs: dict = {"dtype_backend": _default_dtype_backend},
                excel_kwargs: dict = {"dtype_backend": _default_dtype_backend},
                rm_common_prefix: bool = True
                ) -> _pd.DataFrame | dict[str, _pd.DataFrame]:
    '''Import data

    Can be used to import data from one or several filepaths

    Parameters
    ----------
    fpaths: string, Path, or sequence of
        file paths (supported formats: .csv .xls .xlsx .zip .csv.gpg .xlsx.gpg)
    csv_kwargs: dict
        parameter passed to read_csv
    excel_kwargs: dict
        parameter passed to read_excel
    rm_common_prefix: bool
        if dataset share the same common prefix, remove it

    Returns
    -------
    A dict of DataFrame

    '''
    # uniform 1 to many and clean input
    if isinstance(fpaths, str) or isinstance(fpaths, _Path):
        fpaths = [fpaths]

    rval: dict[str, _pd.DataFrame] = {}

    for fpath in fpaths:
        # breakpoint()
        # normalize to Path
        fpath = _Path(fpath)
        if not fpath.exists():
            msg = f"{fpath} does not exists. Ignored"
            _logging.warning(msg)
            continue
        fext = "".join(fpath.suffixes)
        fname = fpath.name.replace(fext, "")
        if fext in {".csv", ".csv.gpg"}:
            data = _import_csv(fpath, csv_kwargs)
            dfname = fname
            if dfname not in rval.keys():  # check for duplicates
                rval[dfname] = data
            else:
                msg = f"{dfname} is duplicated, skipping to avoid overwriting"
                _logging.warning(msg)
        elif fext in {".xls", ".xlsx", ".xls.gpg", ".xlsx.gpg"}:
            excel_sheets = _import_excel(fpath, excel_kwargs)
            rval.update(excel_sheets)
        elif fext == ".zip":  # unzip in temdir and go by recursion
            with _tempfile.TemporaryDirectory() as tempdir:
                with _zipfile.ZipFile(fpath) as myzip:
                    myzip.extractall(tempdir)
                    zipped_fpaths = [
                        _os.path.join(tempdir, f) for f in _os.listdir(tempdir)
                    ]
                    zipped_data = import_data(zipped_fpaths)
            # prepend zip name to fname (as keys) and update results
            zipped_data = {f"{fname}_{k}": v for k, v in zipped_data.items()}
            rval.update(zipped_data)
        else:
            msg = f"File format not supported for {fpath}. Ignoring it."
            _logging.warning(msg)

    if len(rval) == 1:
        # a single dataset: KISS and return the data directly
        for v in rval.values():
            return v
    elif len(rval) > 1:
        # multiple dataset return the dict and remove common name prefix
        if rm_common_prefix:
            keys = list(rval.keys())
            common_prefix = _os.path.commonprefix(keys)
            rval = {k.removeprefix(common_prefix): rval[k]
                    for k in rval.keys()}
        return rval
    else:
        msg = "No data to be imported."
        _logging.error(msg)


# _rdf: pd.DataFrame to R data.frame converter (a-la dput)
# --------------------------------------------------------
def _rdf_missing(x: _pd.Series, xn: str):
    rval = f"{xn} = c(NA)"
    return rval


def _rdf_integer(x: _pd.Series, xn: str):
    data_str = x.to_string(
        na_rep="NA",
        index=False,
        header=False,
    ).replace("\n", ", ")
    rval = f"{xn} = c({data_str})"
    return rval


# placeholder
_rdf_numeric = _rdf_integer


def _rdf_factor(x: _pd.Series, xn: str):
    # to be safe it's seems to be better rather than
    # pd.Categorical
    x = _pd.Series(x, dtype="category")
    # raw data (integers)
    data = x.cat.codes
    data_str = "c({})".format(
        data.to_string(
            na_rep="NA",
            index=False,
            header=False,
        ).replace("\n", ", ")
    )
    # unique categories and labels
    categs = x.cat.categories
    levels = []
    labels = []
    for lev, lab in enumerate(categs):
        levels.append(str(lev))
        labels.append(f'"{lab}"')

    levs = ", ".join(levels)
    labs = ", ".join(labels)
    levels_str = f"levels = c({levs})"
    labels_str = f"labels = c({labs})"
    rval = f"{xn} = factor({data_str}, {levels_str}, {labels_str})"
    return rval


def _rdf_object(x: _pd.Series, xn: str):
    data_l = x.to_list()
    data_l2 = []
    for s in data_l:
        if isinstance(s, str) and s != "":
            data_l2.append(f'"{s}"')
        else:
            data_l2.append("NA")
    data_str = ", ".join(data_l2)
    rval = f"{xn} = c({data_str})"
    return rval


def _rdf_bool(x: _pd.Series, xn: str):
    ft = {True: "TRUE", False: "FALSE"}
    data_str = (
        x.map(ft).to_string(
            na_rep="NA",
            index=False,
            header=False,
        ).replace("\n", ", ")
    )
    rval = f"{xn} = c({data_str})"
    return rval


def _datetimemap(x):
    if _pd.isna(x):
        return "NA"
    else:
        return x.strftime("'%Y-%m-%d %H:%M:%S'")


def _rdf_datetime(x: _pd.Series, xn: str):
    recoded = x.map(_datetimemap)
    val = ", ".join(recoded)
    rval = f"{xn} = c({val})"
    return rval


def _datemap(x):
    if _pd.isna(x):
        return "NA"
    else:
        return x.strftime("'%Y-%m-%d'")


def _rdf_date(x: _pd.Series, xn: str):
    recoded = x.map(_datemap)
    val = ", ".join(recoded)
    rval = f"{xn} = c({val})"
    return rval


def _rdf_datetime_coercion_code(varlist, are_dates=True):
    v2 = [f"'{v}'" for v in varlist]
    v2 = ", ".join(v2)
    if are_dates:
        # dates code
        varlist = f"datevars <- c({v2});"
        importcmd = "df[, datevars] <- lapply(df[, datevars], as.Date);"
        cleaning = "rm(datevars);"
        return [varlist, importcmd, cleaning]
    else:
        # datetimes code
        varlist = f"dt_vars <- c({v2});"
        importfun = "imp_dt <- function(x) " \
            "as.POSIXct(x, format='%Y-%m-%d %H:%M:%S');"
        importcmd = "df[, dt_vars] <- lapply(df[, dt_vars], imp_dt);"
        cleaning = "rm(dt_vars, imp_dt);"
        return [varlist, importfun, importcmd, cleaning, ""]


def _rdf(df: _pd.DataFrame,
         path: str | _Path,
         dfname: str = "df"):
    """
    pd.DataFrame to R data.frame 'converter'
    """
    path = _Path(path)

    r_code = []
    date_vars = []
    datetime_vars = []
    r_code.append(f"{dfname} <- data.frame(")
    for var in df.columns:
        x = df[var]
        if _is_bool(x):
            r_code.append(_rdf_bool(x, var))
        elif _is_integer(x):
            r_code.append(_rdf_integer(x, var))
        elif _is_numeric(x):
            r_code.append(_rdf_numeric(x, var))
        elif _is_categorical(x):
            r_code.append(_rdf_factor(x, var))
        elif _is_date(x):
            r_code.append(_rdf_date(x, var))
            date_vars.append(var)
        elif _is_datetime(x):
            r_code.append(_rdf_datetime(x, var))
            datetime_vars.append(var)
        elif _is_string(x):
            r_code.append(_rdf_object(x, var))
        elif _is_all_missing(x):
            r_code.append(_rdf_missing(x, var))
        else:
            msg = f"{var}: il tipo {x.dtype!r} non è ancora gestito."
            raise ValueError(msg)
        is_last = var == df.columns[-1]
        if is_last:
            r_code.append(")\n")
        else:
            r_code.append(",\n")

    # fix some NA
    r_code = [line.replace("<NA>", "NA") for line in r_code]
    # add code for date/datetime import
    if date_vars:
        r_code += _rdf_datetime_coercion_code(date_vars, are_dates=True)
    if datetime_vars:
        r_code += _rdf_datetime_coercion_code(datetime_vars, are_dates=False)
    # finally export the surce code
    with path.open(mode="w") as f:
        f.writelines(r_code)


def _export_excel(x, xlsx_path, index):
    """
    Export excel file handling for encryption, set using .xslx.gpg in
    export_data

    """
    do_encryption = ".gpg" in xlsx_path.suffixes
    unencrypted_tempfname = _tempfile.mkstemp(suffix='.xlsx')[1]
    # export unencrypted x in temp file
    with _pd.ExcelWriter(unencrypted_tempfname) as writer:
        if isinstance(x, _pd.DataFrame):
            x.to_excel(writer,
                       sheet_name="Foglio 1",
                       index=index)
        elif isinstance(x, dict):
            for k, v in x.items():
                # preprocess dict's key to be accepted as excel sheetname
                k = _fix_varnames(k)[:31]
                v.to_excel(writer, sheet_name=k, index=index)
    # if encrypted do encryption
    if do_encryption:
        # just encrypt the file above in the target destination
        cmd = "gpg --encrypt --recipient lbraglia@gmail.com" \
            f" --output {xlsx_path} {unencrypted_tempfname}"
        _os.system(cmd)
    else:
        # just copy to destination
        _shutil.copy(unencrypted_tempfname, xlsx_path)
    # in any case drop the unencrypted version
    _Path(unencrypted_tempfname).unlink()


def export_data(x: _pd.DataFrame | dict[str, _pd.DataFrame],
                path: str | _Path,
                ext: str | list[str] = [".xlsx", ".csv",
                                        ".pkl", ".R", ".feather"],
                index=False,
                dfname="df"
                ) -> None:
    """Export a DataFrame or a dict of DataFrames as csv/xlsx

    In case of a dict is used and a csv path is given, the path is
    suffixed with dict names

    Parameters
    ----------
    x:
        dict of pandas.DataFrame
    fpath:
        file path, if extension is provided it will overwrite
        formats otherwise formats is considered
    ext:
        str or list of string with file extensions
    index:
        bool add index in exporting (typically True for results,
        False for data)
    verbose: bool
        for single variables exporting (R) say the varname which is processed
    """

    if not (isinstance(x, _pd.DataFrame) or isinstance(x, dict)):
        msg = "x deve essere un pd.DataFrame o un dict di pd.DataFrame"
        raise ValueError(msg)

    if isinstance(ext, str):  # uniforma
        ext = [ext]

    path = _Path(path)
    path_has_suffix = path.suffix != ""
    if path_has_suffix:
        # used_formats = [path.suffix.replace(".", "")]
        used_formats = ["".join(path.suffixes)]
    else:
        used_formats = ext

    # don't put ".xlsx" and ".xlsx.gpg" in the same branch otherwise only one
    # get exported
    if ".xlsx" in used_formats:
        xlsx_path = path if path_has_suffix else path.with_suffix(".xlsx")
        _export_excel(x=x, xlsx_path=xlsx_path, index=index)

    if ".xlsx.gpg" in used_formats:
        xlsx_path = path if path_has_suffix else path.with_suffix(".xlsx.gpg")
        _export_excel(x=x, xlsx_path=xlsx_path, index=index)

    if ".csv" in used_formats:
        if isinstance(x, _pd.DataFrame):
            x.to_csv(path if path_has_suffix else path.with_suffix(".csv"),
                     index=index)
        elif isinstance(x, dict):
            # use dict key as postfix
            for k, v in x.items():
                csv_path = path.parent / (str(path.stem) + f"_{k}.csv")
                v.to_csv(csv_path, index=index)

    if ".pkl" in used_formats:
        if isinstance(x, _pd.DataFrame):
            x.to_pickle(path if path_has_suffix else path.with_suffix(".pkl"))
        elif isinstance(x, dict):
            # use dict key as postfix
            for k, v in x.items():
                pkl_path = path.parent / (str(path.stem) + f"_{k}.pkl")
                v.to_pickle(pkl_path)

    if ".R" in used_formats:
        if isinstance(x, _pd.DataFrame):
            _rdf(x,
                 path if path_has_suffix else path.with_suffix(".R"),
                 dfname)
        elif isinstance(x, dict):
            # use dict key as postfix
            for k, v in x.items():
                r_path = path.parent / (str(path.stem) + f"_{k}.R")
                _rdf(v, r_path, dfname=k)

    if ".feather" in used_formats:
        if isinstance(x, _pd.DataFrame):
            x.to_feather(path if path_has_suffix
                         else path.with_suffix(".feather"))
        elif isinstance(x, dict):
            # use dict key as postfix
            for k, v in x.items():
                feather_path = path.parent / (str(path.stem) + f"_{k}.feather")
                v.to_feather(feather_path)


# ------------------------------------
# LaTeX stuff
# ------------------------------------
# Thanks PyLaTeX guys
_latex_special_chars = {
    "&": r"\&",
    "%": r"\%",
    "$": r"\$",
    "#": r"\#",
    "_": r"\_",
    "{": r"\{",
    "}": r"\}",
    "~": r"\textasciitilde{}",
    "^": r"\^{}",
    "\\": r"\textbackslash{}",
    "\n": "\\newline%\n",
    "-": r"{-}",
    "\xA0": "~",  # Non-breaking space
    "[": r"{[}",
    "]": r"{]}",
}


def _latex_escape(s):
    """
    latex_escape from PyLaTeX

    s: a str or something coercible to it
    >>> print(latex_escape("asd_foo_bar"))
    """
    string = str(s)
    return "".join(_latex_special_chars.get(c, c) for c in string)


def latex_table(
    tab,
    label: str = "",
    caption: str = "",
    position: str | None = None,
    float_format: str = "%.1f",
    column_format: str | None = None,
):
    """Print a LaTeX table

    Print a pd.DataFrame or inheriting object with to_latex method with
    sensible defaults.

    Parameters
    ----------
    tab:
        something with .to_latex method
    label: str
        posta dopo "fig:" in LaTeX
    caption: str
        caption tabella LaTeX
    position: str
        lettere posizionamento tabella (ad esempio
        https://stackoverflow.com/questions/1673942)
    """
    if (label == "") or (not isinstance(label, str)):
        msg = "Please provide a label for the table."
        raise ValueError(msg)
    caption = (
        label.capitalize().replace("_", " ") if caption == "" else caption
    )
    latex_caption = _latex_escape(caption)
    latex_label = "tab:" + label
    if isinstance(tab, _pd.DataFrame):
        tab.columns.name = None
    if (column_format is None) and isinstance(tab, _pd.DataFrame):
        ncols = tab.shape[1]
        column_format = "".join(["l"] + ["r"] * ncols)
    # per avere il centering è necessario impostare lo stile
    # https://pandas.pydata.org/docs/reference/api/pandas.io.formats.style.Styler.to_latex.html
    content = tab.to_latex(
        # fissi
        na_rep="",
        index=True,
        index_names=False,
        escape=True,
        # position_float='centering', # one day. maybe.
        # variabili
        label=latex_label,
        caption=latex_caption,
        position=position,
        float_format=float_format,
        column_format=column_format,
    )
    print(content)



def _md_to_str(x):
    # Funzione di utilita per convertire ciascuna cella di un dataframe ad una
    # stringa utile per la stampa di tabella markdown
    if _pd.isna(x) or x == "":
        return ""
    elif isinstance(x, float):
        return str(int(round(x, 0)))
    else:
        return str(x)



def md_table(
    tab,
    label: str = "",
    caption: str = "",
    position: str | None = None,
    float_format: str = "%.1f",
    column_format: str | None = None,
    **kwargs
):
    """Print a Markdown table (using pandas)

    Print a pd.DataFrame or inheriting object with to_latex method with
    sensible defaults.

    Parameters
    ----------
    tab:
        something with .to_markdown method
    label: str
        posta dopo "fig:" in LaTeX
    caption: str
        caption tabella LaTeX
    position: str
        lettere posizionamento tabella (ad esempio
        https://stackoverflow.com/questions/1673942)
    kwargs: dict
        altri parametri passati a .to_markdown()
    """
    # if (label == "") or (not isinstance(label, str)):
    #     msg = "Please provide a label for the table."
    #     raise ValueError(msg)
    # caption = (
    #     label.capitalize().replace("_", " ") if caption == "" else caption
    # )
    # latex_caption = _latex_escape(caption)
    # latex_label = "tab:" + label
    # if isinstance(tab, _pd.DataFrame):
    #     tab.columns.name = None
    # if (column_format is None) and isinstance(tab, _pd.DataFrame):
    #     ncols = tab.shape[1]
    #     column_format = "".join(["l"] + ["r"] * ncols)
    # # per avere il centering è necessario impostare lo stile
    # #
    # https://pandas.pydata.org/docs/reference/api/pandas.io.formats.style.Styler.to_latex.html
    content = tab.map(_md_to_str).to_markdown(
        # fissi
        tablefmt="pipe",
        # na_rep="",
        # index=True,
        # index_names=False,
        # escape=True,
        # # position_float='centering', # one day. maybe.
        # # variabili
        # label=latex_label,
        # caption=latex_caption,
        # position=position,
        # float_format=float_format,
        # column_format=column_format,
        **kwargs
    )
    print(content)


def export_tables(tabs_dict: dict[str, _pd.DataFrame],
                  outfile: str | _Path = "outputs/tables.xlsx"):
    """Latex print and excel Export a dict of tables (caption as key)

    Parameters
    ----------
    tabs_dict:
       dict of DataFrame to be exported
    outfile:
       path to the excel file for export

    Examples
    --------
    >>> exported = {"Descrittive": desc_df, "Analisi 1"  : analysis1_df,
    >>>             "Analisi 2"  : analysis2_df}
    >>> export_tables(exported)

    """
    # excel
    export_data(tabs_dict, outfile)
    # latex
    labels = []
    for caption, tab in tabs_dict.items():
        lab = _fix_varnames(caption)
        labels.append(lab)
        latex_table(tab, label=lab, caption=caption)
    # return latex references
    refs = ["\ref{tab:" + lab + "}" for lab in labels]
    return refs


# functions to hide certain section conditioned on a logical values
def hide_if(condition, markup="md"):
    markup = _match_arg(markup, ["tex", "md"])
    if condition:
        what = b"\\iffalse" if markup == 'tex' else "<!--"
        print(what)


def hide_fi(condition, markup="md"):
    markup = _match_arg(markup, ["tex", "md"])
    if condition:
        what = b"\\fi" if markup == 'tex' else "-->"
        print(what)



""" Utilities coming from R."""

import itertools as _itertools
import inspect as _inspect
import numpy as _np
import pandas as _pd
import types as _types
import subprocess as _subprocess
import tempfile as _tempfile
from pprint import pformat as _pformat
import re as _re


def match_arg(arg, choices):
    """R's match.arg equivalent for programming function for interactive use

    Examples
    --------
    >>> from pylbmisc.r import match_arg
    >>> # questo ritorna errore perché matcha troppo
    >>> user_input = "foo"
    >>> # a = match_arg(user_input, ["foobar", "foos", "asdomar"]) # errore
    >>> # questo è ok e viene espanso
    >>> user_input2 = "foob"
    >>> a = match_arg(user_input2, ["foobar", "foos", "asdomar"])
    >>> print(a)
    foobar
    """
    res = [expanded for expanded in choices if expanded.startswith(arg)]
    choices_str = ", ".join(choices)
    res_len = len(res)
    if res_len == 0:
        msg = f"Parameter {arg} must be one of: {choices_str}"
        raise ValueError(msg)
    elif res_len > 1:
        msg = f"Parameter {arg} matches multiple choices from: {choices_str}"
        raise ValueError(msg)
    else:
        return res[0]


def expand_grid(dictionary):
    # https://stackoverflow.com/questions/12130883
    """Replacement for R's expand.grid

    Examples
    --------
    >>> import pylbmisc as lb
    >>> stratas =  {"centres": ["ausl re", "ausl mo"],
    ...             "agecl": ["<18", "18-65", ">65"],
    ...             "foo": ["1"]}
    >>> lb.r.expand_grid(stratas)
       centres  agecl foo
    0  ausl re    <18   1
    1  ausl re  18-65   1
    2  ausl re    >65   1
    3  ausl mo    <18   1
    4  ausl mo  18-65   1
    5  ausl mo    >65   1
    """
    rows = [row for row in _itertools.product(*dictionary.values())]
    return _pd.DataFrame(rows, columns=dictionary.keys())


def dput(x) -> None:
    """Try to print the ASCII representation of a certain object

    Parameters
    ----------
    x : anything
        data to be printed

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from pylbmisc.r import dput
    >>> import pylbmisc as lb
    >>> List = list(range(0, 15))
    >>> dput(List)
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
    >>> Dict = {"a": 1, "b": 2}
    >>> dput(Dict)
    {'a': 1, 'b': 2}
    >>> Dict2 = {"a": List, "b": List}
    >>> dput(Dict2)
    {'a': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14],
     'b': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]}
    >>> nparray = np.array(List)
    >>> dput(nparray)
    np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])
    >>> series = pd.Series(List)
    >>> dput(series)
    pd.Series([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14])
    >>> def function() -> None:
    ...     pass
    >>> dput(function)
    def function() -> None:
        pass
    <BLANKLINE>
    >>> df = lb.datasets.load()
    >>> dput(df)
    pd.DataFrame({'dead': [6, 13, 18, 28, 52, 53, 61, 60],
     'logdose': [1.691, 1.724, 1.755, 1.784, 1.811, 1.837, 1.861, 1.884],
     'n': [59, 60, 62, 56, 63, 59, 62, 60]})
    """
    if isinstance(x, _types.FunctionType):  # special cases: don't use repr
        print(_inspect.getsource(x))
    elif isinstance(x, _pd.DataFrame):
        dict_rep = _pformat(x.to_dict(orient="list"), compact=True)
        print(f"pd.DataFrame({dict_rep})")
    elif isinstance(x, _pd.Series):
        list_rep = _pformat(x.to_list(), compact=True)
        print(f"pd.Series({list_rep})")
    elif isinstance(x, _np.ndarray):
        list_rep = _pformat(x.tolist(), compact=True)
        print(f"np.array({list_rep})")
    else:
        obj_repr = _pformat(x, compact=True)
        print(obj_repr)


def table(x: _pd.Series | None = None,
          y: _pd.Series | None = None,
          **kwargs):
    """Emulate the good old table for quick crosstabs.

    Parameters
    ----------
    x: pd.Series
       first variable
    y: pd.Series
       second variable
    kwargs: Any
       other parameters passed to Series.value_counts or pd.crosstab
    """
    if x is None:
        msg = "Almeno una variable"
        raise ValueError(msg)
    if y is None:
        return x.value_counts(dropna=False, **kwargs)
    else:
        return _pd.crosstab(x, y, dropna=False, margins=True, **kwargs)


def debug(step_into_this):
    r"""Emulate R's debug function as a decorator

    Parameters
    ----------
    step_into_this: function
        the function to be breakpoint()-ed

    Examples
    --------
    >>> # comment the decorator when done with debugging
    >>> @debug
    >>> def myfunction(x, y=1):
    ...     m = 1
    ...     return x * y * m
    >>>
    >>> myfunction(5) # this call is breakpointed
    """
    def wrapper(*args, **kwargs):
        breakpoint()
        return step_into_this(*args, **kwargs)
    return wrapper


def view(df: _pd.DataFrame | _pd.Series) -> None:
    """View a pd.DataFrame using LibreOffice.

    Parameters
    ----------
    df:
        the dataframe to be visualized

    Examples
    --------
    >>> # lb.dm.view(df) # commented to avoid tests fails
    """
    if not isinstance(df, (_pd.DataFrame, _pd.Series)):
        msg = "Only dataframes are visualized."
        raise Exception(msg)
    tempfile = _tempfile.mkstemp(suffix=".xlsx")
    fname = tempfile[1]
    df.to_excel(fname)
    _subprocess.Popen(["libreoffice", fname])


_comment_re = _re.compile("(#.+)")
_prompt_re = _re.compile("^(>{3} ?|\\.{3} ?)")


def example(f):
    """Extract example code to be executed

    Parameters
    ----------
    f: function
       function for doc's example

    Examples
    --------
    >>> import pylbmisc as lb
    >>> exec(lb.r.example(lb.stats.p_adjust))
    """
    doc = _inspect.getdoc(f)
    # select lines with >>> or ...
    code = [line for line in doc.split("\n")
            if line.startswith(">>>")
            or line.startswith("...")]
    print("\n".join(code))
    # rm from comments onward
    uncommented = [_comment_re.sub("", line) for line in code]
    # rm starting lines
    clean_code = [_prompt_re.sub("", line) for line in uncommented]
    to_be_run = [line for line in clean_code if line != ""]
    return "\n".join(to_be_run)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""Miscellaneous utilities

This module has utilities for everyday work such UI/UX (argument parsing,
interactive ascii menu), system info
"""

import argparse as _argparse
import datetime as _dt
import inspect as _inspect
import re as _re
# import readline as _readline # be more portable
import sys as _sys
import time as _time
import types as _types
import importlib as _importlib

from pprint import pp as _pp
from pylbmisc.iter import unique as _unique
from typing import Sequence as _Sequence


# _readline.parse_and_bind("set editing-mode emacs") # hopefully


def is_interactive():
    """Check if python is running in interactive mode

    Returns
    -------
    bool: True if running in interactive mode, False if running using
         ipykernel/quarto

    Examples
    --------
    >>> import sys
    >>> from pyblmisc.utils import is_interactive
    >>> is_interactive()
    False  # tests aren't run interactively
    """
    parent_frame = _inspect.currentframe().f_back
    test = "'__PYTHON_EL_eval' in dir()"  # active in Emacs, not in quarto
    return eval(test, parent_frame.f_locals, parent_frame.f_globals)


def argparser(opts):
    """Helper function for argument parsing.

    Examples
    --------
    >>> import pylbmisc as lb # third party
    >>>
    >>> opts = (
    ...  # (param, help, default, type)
    ...  # --dirs
    ...  ("dirs", "str: comma separated list of exercise source directories",
    ...  "~/src/pypkg/exercises/db", str),
    ...  # --list
    ...  ("list", "str: comma separated list of file with lists of source dir",
    ...   None, str),
    ...  # --outfile
    ...  ("outfile", "str:  sqlite3 db to save", "~/.exercises.db", str))
    >>> # args = lb.utils.argparser(opts)   # tests, you know ...
    >>> # dirs = args["dirs"]
    >>> # dirs = dirs.split(",")
    >>> # lists = args["lists"]
    >>> # lists = lists.split(",")
    >>> # outfile = args["outfile"]
    >>> # print({"dirs": dirs, "lists": lists, "outfile": outfile})
    """
    parser = _argparse.ArgumentParser()
    # defaults = {}
    for i in opts:
        optname = i[0]
        optdescription = i[1]
        optdefault = i[2]
        opttype = i[3]
        # create help string and add argument to parsing
        help_string = f"{optdescription} (default: {optdefault!r})"
        parser.add_argument("--" + optname, help=help_string, type=str)
    # do parsing
    args = vars(parser.parse_args())  # vars to change to a dict
    # defaults settings and types management
    for i in opts:
        optname = i[0]
        optdescription = i[1]
        optdefault = i[2]
        opttype = i[3]
        # se il valore è a none in args impostalo al valore di default
        # specificato
        if args[optname] is None:
            args[optname] = optdefault
        # se il tipo è logico sostituisci un valore possibile true con
        # l'equivalente python
        if opttype is bool:
            # mv to character if not already (not if used optdefault)
            args[optname] = str(args[optname])
            true_values = (
                "true",
                "True",
                "TRUE",
                "t",
                "T",
                "1",
                "y",
                "Y",
                "yes",
                "Yes",
                "YES",
            )
            if args[optname] in true_values:
                args[optname] = "True"
            else:
                args[optname] = ""
        # converti il tipo a quello specificato, a meno che non sia None se no
        # lascialo così
        if args[optname] is not None:
            args[optname] = opttype(args[optname])
    return args


def line_to_numbers(x: str) -> list[int]:
    """Transform a string of positive numbers "1 2-3, 4, 6-10" into a
    list [1,2,3,4,6,7,8,9,10]

    Parameters
    ----------
    x: str
       string containing numbers such as printers pages lists

    Examples
    --------
    >>> from pylbmisc.utils import line_to_numbers
    >>> line_to_numbers("1, 2-4, 16-18")
    [1, 2, 3, 4, 16, 17, 18]
    """
    # # replace comma with white chars
    x = x.replace(",", " ")
    # keep only digits, - and white spaces
    x = _re.sub(r"[^\d\- ]", "", x)
    # split by whitespaces
    spl = x.split(" ")
    # change ranges to proper
    expanded = []
    single_page_re = _re.compile("^[0-9]+$")
    pages_range_re = _re.compile("^([0-9]+)-([0-9]+)$")
    for i in range(len(spl)):
        # Check if the single element match one of the regular expression
        single_page = single_page_re.match(spl[i])
        pages_range = pages_range_re.match(spl[i])
        if single_page:
            # A) One single page: append it to results
            expanded.append(spl[i])
        elif pages_range:
            # B) Pages range: append a list of (expanded) pages to results
            first = int(pages_range.group(1))
            second = int(pages_range.group(2))
            # step is 1 if first is less than or equal to second or -1
            # otherwise
            step = 1 * int(first <= second) - 1 * int(first > second)
            if step == 1:
                second += 1
            elif step == -1:
                second -= 1
            else:
                # do nothing (ignore if they don't match)
                pass
            expanded_range = [str(val) for val in range(first, second, step)]
            expanded += expanded_range
    # coerce to integer expanded
    res: list[int] = [int(x) for x in expanded]
    return res


def menu(choices: _Sequence[str],
         title: str | None = None,
         allow_repetition: bool = False,
         strict: bool = True,
         ) -> _Sequence[str]:
    """
    CLI menu for user single/multiple choices

    Parameters
    ----------
    choices: list of string
         printed choices
    title: str
         a pretty printed title
    allow_repetition: bool
         multiple selected items
    strict: bool
         all the selection number are among the available ones
         (or take the consistent otherwise)
    """
    available_ind = [i + 1 for i in range(len(choices))]
    avail_with_0 = [0, *available_ind]
    the_menu = "\n".join([str(i) + ". " + str(c)
                          for i, c in zip(available_ind, choices)])
    if title:
        ascii_header(title)
    print(the_menu, "\n")
    select_msg = "Selection (values as '1, 2-3, 6') or 0 to exit: "
    ind = line_to_numbers(input(select_msg))
    # normalize to list (for single selections, for now)
    if not isinstance(ind, list):
        ind = list(ind)
    if strict:
        # continue asking for input until all index are between the selectable
        while not all(i in avail_with_0 for i in ind):
            not_in = [i for i in ind if i not in avail_with_0]
            print("Not valid insertion: ", not_in, "\n")
            ind = line_to_numbers(input(select_msg))
            if not isinstance(ind, list):
                ind = list(ind)
    else:
        # keep only the input in avail_with_0
        allowed = [i for i in ind if i in avail_with_0]
        any_not_allowed = not all(allowed)
        if any_not_allowed:
            print(
                "Removed some values (not 0 or specified possibilities): ",
                list(set(ind) - set(allowed)),
                ".",
            )
            ind = allowed
    # make unique if not allowed repetitions
    if not allow_repetition:
        ind = list(_unique(ind))
    # obtain the selection: return always a list should simplify the code
    rval = [choices[i - 1] for i in ind if i != 0]
    return rval


def ascii_header(x: str) -> None:
    """
    Create an ascii header given a string as title.

    Parameters
    ----------
    x: str
        message to be prettyprinted

    Examples
    --------
    >>> import pylbmisc as lb
    >>> ascii_header("Hello, world!")
    =============
    Hello, world!
    =============
    <BLANKLINE>
    """
    header = "=" * len(x)
    msg = f"{header}\n{x}\n{header}\n"
    print(msg)


def _imports(globs):
    "Something like https://stackoverflow.com/questions/4858100"
    for val in globs.values():
        if isinstance(val, _types.ModuleType):
            yield val


def sysinfo():
    """Display several system informations.

    The aim should be to emulate R's sessionInfo + additional uv/venv info.
    
    Examples
    --------
    >>> import pylbmisc as lb
    >>> lb.utils.sysinfo()
    """
    parent_frame = _inspect.currentframe().f_back
    globs = parent_frame.f_globals
    _pp({
        "python-path": _sys.executable,
        "python-version": _sys.version,
        "imports": list(_unique(_imports(globs=globs))),
        "paths": _sys.path,
        "now": _dt.datetime.now().isoformat(),
    })


def timeit(n=1):
    """A timer decorator with parameter for number of execution

    Parameters
    ----------
    n: integer
       number of function execution (default=1)

    Examples
    --------
    >>> import pylbmisc as lb
    >>> import time
    >>> @timeit
    >>> def test():
    ...      time.sleep(2)
    >>>
    >>> test()
    >>>
    >>> @timeit(n=3)
    >>> def test2():
    ...      time.sleep(2)
    >>>
    >>> test2()
    """
    def inner(f):
        def wrapper(*args, **kwargs):
            start_time = _time.perf_counter()
            for i in range(n):
                rval = f(*args, **kwargs)
            stop_time = _time.perf_counter()
            elapsed = stop_time - start_time
            function_name = f.__name__
            msg_start = f"{n} calls of" if n > 1 else "Function"
            msg_end = f" (mean: {elapsed/n:.4f} secs)" if n > 1 else "."
            msg = f"{msg_start} {function_name} [{args=} {kwargs=}] " \
                f"took {elapsed:.2f} secs{msg_end}"
            print(msg)
            return rval
        return wrapper
    return inner



def namespace(pkgname="pylbmisc"):
    """This should print a tree of module, submodule and contents

    Parameters
    ----------
    pkgname: str
        name of the package to be printed

    Examples
    --------
    >>> # namespace("pylbmisc")
    >>> # namespace("scipy") # TODO evolve the function to handle this
    """
    # da riscrivere come funzione ricorsiva, qui fa solo il primo livello, che
    # per pylbmisc è abbastanza per avere una idea della roba esportata
    mod = _importlib.import_module(pkgname)
    modules = [m for m in dir(mod) if not m.startswith("__")]
    for module_name in modules:
        pre = post = "-" * 30
        print(f"\n{pre} {module_name} {post}\n")
        mod = _importlib.import_module("." + module_name, package=pkgname)
        objects = [(name, getattr(mod, name))
                   for name in getattr(mod, "__all__", dir(mod))
                   if not name.startswith("_")]
        for x in objects:
            _pp(x)


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import matplotlib.pyplot as _plt
import numpy as _np
import pandas as _pd
from collections.abc import Callable as _Callable
from pathlib import Path as _Path
from matplotlib.figure import Figure as _Figure
from matplotlib.axes import Axes as _Axes
from typing import Optional as _Optional
from typing import Tuple as _Tuple
from pylbmisc.stats import p_format as _p_format


def fun2d(f: _Callable = lambda x: x**2,
          xlim: list[int | float] = [-5, 5],
          ylim: _Optional[list[int | float]] = None,
          npoints: int = 1000,
          show: bool = True,
          save: _Optional[str | _Path] = None,
          fig: _Optional[_Figure] = None,
          ax: _Optional[_Axes] = None
          ) -> _Tuple[_Figure, _Axes]:
    """Plot a 2d function

    Parameters
    ----------
    f : Callable
        function to be applied to x
    xlim: list[int | float]
        range of x
    ylim: list[int | float] | None
        range of y plotted
    npoints: int
        number of points to plot
    show: bool
        plot at the end of figure composition
    fig: matplotlib.figure.Figure | None
        fig used for plotting or if None a new will be created
    ax: matplotlib.axes.Axes | None
        ax used for plotting or if None a new will be created

    Returns
    -------
    fig: matplotlib.figure.Figure
        fig used for plotting
    ax: matplotlib.axes.Axes
        ax used for plotting

    Examples
    --------
    >>> fig, ax = fun2d()

    """
    if (ax is None) or (fig is None):
        fig, ax = _plt.subplots()
    x = _np.linspace(start=xlim[0], stop=xlim[1], num=npoints)
    npf = _np.frompyfunc(f, 1, 1)
    y = npf(x)
    ax.plot(x, y)
    if ylim is not None:
        ax.set_ylim(bottom=ylim[0], top=ylim[1])
    if show:
        _plt.show(block=False)
    if save is not None:
        fig.savefig(save)
    return fig, ax


def fun3d(f: _Callable = lambda x, y: x**2 + y**2,
          xlim: list[int | float] = [-5, 5],
          ylim: list[int | float] = [-5, 5],
          zlim: _Optional[list[int | float]] = None,
          npoints: int = 100,
          show: bool = True,
          save: _Optional[str | _Path] = None,
          fig: _Optional[_Figure] = None,
          ax: _Optional[_Axes] = None
          ) -> _Tuple[_Figure, _Axes]:
    """Plot a 2d function

    Parameters
    ----------
    f : Callable
        function to be applied to x
    xlim: list[int | float]
        range of x
    ylim: list[int | float]
        range of y
    zlim: list[int | float] | None
        range of z plotted
    npoints: int
        number of points to plot
    show: bool
        plot at the end of figure composition
    fig: matplotlib.figure.Figure | None
        fig used for plotting or if None a new will be created
    ax: matplotlib.axes.Axes | None
        ax used for plotting or if None a new will be created

    Returns
    -------
    fig: matplotlib.figure.Figure
        fig used for plotting
    ax: matplotlib.axes.Axes
        ax used for plotting

    Examples
    --------
    >>> fig, ax = fun3d()

    """
    if (ax is None) or (fig is None):
        fig, ax = _plt.subplots(subplot_kw={"projection": "3d"})
    x = _np.arange(start=xlim[0],
                   stop=xlim[1],
                   step=(xlim[1] - xlim[0])/(npoints - 1))
    y = _np.arange(start=ylim[0],
                   stop=ylim[1],
                   step=(ylim[1] - ylim[0])/(npoints - 1))
    x, y = _np.meshgrid(x, y)
    npf = _np.frompyfunc(f, nin=2, nout=1)
    z = npf(x, y)
    ax.plot_surface(x, y, z)
    if zlim is not None:
        ax.set_zlim(bottom=zlim[0], top=zlim[1])
    if show:
        _plt.show(block=False)
    if save is not None:
        fig.savefig(save)
    return fig, ax


def forestplot(
    df,
    variable_col="variable",
    group_col="group",
    est_col="HR",
    ll_col="lower.95",
    hl_col="upper.95",
    n_col="n",
    pval_col="p",
    ci_digits=2,
    fontfamily="DejaVu Sans",
    fontsize=15,
    forest_xlim=(10**(-1.5), 10**1.5),
    log_scale=True,
    xlabel="Hazard ratios",
    forest_title="Overall survival",
    refline=1,
    refline_color="red",
    refline_style="--",
    dot_color="black",
    ci_color="black",
    row_height=0.55
):
    """
    Examples
    --------
    >>> import pylbmisc as lb
    >>> fig, ax = fun3d()
    >>> import numpy as np
    >>> nan = np.nan
    >>> inf = np.inf
    >>> os_flat = pd.DataFrame({'HR': [1.0, 0.8078596501254128, 1.0, 0.9276629617388676, 1.083148102276055,
    >>>         1.410879356833972, 1.0, 0.9132984204206765, 0.690306667065486,
    >>>         3.619033273498864, 1.0, 1.182346896367745, 1.928877100106704, 1.0,
    >>>         0.2495388668068113, 0.7802993699197853, 4.819819623454819e-07,
    >>>         5.618010353736118, 1.0, 2.761579895556096, 1.0, 2.251540997105791, 1.0,
    >>>         0.6317149418796589, 1.391470021361166, 1.0, 2.450908443888874,
    >>>         1.845714125925274, 2.734585278092313, 1.0, 3.566961223209396, 1.0,
    >>>         7.047547494513447, 2.686379735217062, 6.034625373539202,
    >>>         3.693700054306441],
    >>>  'group': ['male', 'female', '(48.3, 65.3]', '(65.3, 71.9]', '(71.9, 76.4]',
    >>>            '(76.4, 85.9]', 'IgG', 'IgA', 'LC', 'Other', '1', '2', '3+', 'Rd',
    >>>            'VTd', 'Vd', 'VMP', 'Pd', 'no HiR features', 'HiR features', 'no',
    >>>            'yes', '0', '1-2', '3', '(0.8, 4.4]', '(4.4, 5.3]', '(5.3, 7.8]',
    >>>            '(7.8, 30.0]', '1-2', '3', 'SR', 'Non-1q HiRCAs', '+1q',
    >>>            '+1q+HiRCAs', None],
    >>>  'lower.95': [nan, 0.459753262045847, nan, 0.4080223189097797,
    >>>               0.4752521515046189, 0.6322469140653059, nan, 0.4310621390173041,
    >>>               0.3021957023990133, 1.091776016753074, nan, 0.5941364676239838,
    >>>               0.8447511914413539, nan, 0.03396574783737238, 0.3273138267395636,
    >>>               0.0, 2.16272156649981, nan, 1.314557817871279, nan,
    >>>               1.269892148793646, nan, 0.1751051111892889, 0.5327921539120398,
    >>>               nan, 0.5733200703278137, 0.4127226251463592, 0.7003493239886236,
    >>>               nan, 1.71806312084743, nan, 1.979165276456984, 0.7205870869270733,
    >>>               1.797006366220221, 1.285899707849024],
    >>>  'n': [88, 86, 44, 43, 43, 44, 101, 40, 28, 5, 80, 73, 21, 125, 21, 19, 2, 7,
    >>>        154, 20, 120, 54, 27, 29, 54, 26, 16, 19, 20, 97, 35, 43, 11, 18, 20,
    >>>        82],
    >>>  'p': [nan, 0.4581654702047059, nan, 0.8577977636500718, 0.8492832620467534,
    >>>        0.4006380332916023, nan, 0.8128508979186799, 0.3792091009044879,
    >>>        0.035416084331664, nan, 0.6333101510801474, 0.1188861421310271, nan,
    >>>        0.1724836065789395, 0.5757003918594448, 0.9957899407164232,
    >>>        0.0003945510104673111, nan, 0.007315940715285618, nan,
    >>>        0.005474671508132051, nan, 0.4829015839460461, 0.50000305574803, nan,
    >>>        0.226496794047904, 0.4225825089461596, 0.1477644150842377, nan,
    >>>        0.000644908637339989, nan, 0.00258237442308853, 0.1410526685080054,
    >>>        0.003634264049804077, 0.01522236870908805],
    >>>  'upper.95': [nan, 1.419537974339969, nan, 2.109096808433195, 2.468604945290432,
    >>>               3.14842273684017, nan, 1.935020335222294, 1.576869858876642,
    >>>               11.9964183437949, nan, 2.352900822501218, 4.40433456029561, nan,
    >>>               1.833307081751119, 1.86019366417318, inf, 14.59366791526791, nan,
    >>>               5.801436358188687, nan, 3.992021579520693, nan, 2.278995542069771,
    >>>               3.63404154158489, nan, 10.47748458708434, 8.254116510894042,
    >>>               10.67746678267756, nan, 7.405555834062594, nan, 25.09539060645614,
    >>>               10.01493950239945, 20.26520555715178, 10.61001881243547],
    >>>  'variable': ['Sex', nan, 'Age', nan, nan, nan, 'Paraprotein', nan, nan, nan,
    >>>               'Line of therapy', nan, nan, 'Daratumumab combination', nan, nan,
    >>>               nan, nan, 'HiR clinical features', nan, 'Renal insufficiency',
    >>>               nan, 'PET/CT lesions', nan, nan, 'PET/CT SUVmax', nan, nan, nan,
    >>>               'ISS stage', nan, 'Cytogenetic status', nan, nan, nan, nan]})
    >>>
    >>>
    >>> fp_os = lb.fig.forestplot(os_flat, forest_title = "OS")
    >>> lb.io.export_figure(fp_os, fdir="/tmp",fname = "fp_os")
    """
    # breakpoint()
    df = df.copy()
    n_rows = len(df)
    # replace NA with string NA for groups of blank for variables
    df.loc[df[variable_col].isna(), variable_col] = ""
    df.loc[df[group_col].isna(), group_col] = "<NA>"
    # graph parameters
    len_longest_variable = df[variable_col].str.len().max()
    variable_col_width = 0.19 * len_longest_variable
    len_longest_group = df[group_col].str.len().max()
    group_col_width = 0.19 * len_longest_group
    n_col_width = 1.3
    ci_col_width = 2.5
    pval_col_width = 1.4
    forest_col_width = 5
    width_ratios = [variable_col_width, group_col_width, n_col_width,
                    ci_col_width, pval_col_width, forest_col_width]
    fig_width = sum(width_ratios) + 1
    fig_height = row_height * n_rows + 1.5

    def ci_format(row):
        est = row[est_col]
        ll = row[ll_col]
        hl = row[hl_col]
        if _pd.isna(ll) and _pd.isna(hl):  # reference group with lower and upper
            rval = f"{est:.{ci_digits}f}"
        else:
            rval = f"{est:.{ci_digits}f} ({ll:.{ci_digits}f}-{hl:.{ci_digits}f})"
        return rval
    ci_texts = df.apply(ci_format, axis=1)
    pval_texts = _p_format(df[pval_col])
    n_texts = df[n_col].astype(int).astype(str)

    # figure setup
    fig = _plt.figure(figsize=(fig_width, fig_height))
    gs = fig.add_gridspec(1, 6, width_ratios=width_ratios, wspace=0.1)
    ax_variables = fig.add_subplot(gs[0, 0])
    ax_groups = fig.add_subplot(gs[0, 1], sharey=ax_variables)
    ax_n = fig.add_subplot(gs[0, 2], sharey=ax_variables)
    ax_ci = fig.add_subplot(gs[0, 3], sharey=ax_variables)
    ax_pval = fig.add_subplot(gs[0, 4], sharey=ax_variables)
    ax_forest = fig.add_subplot(gs[0, 5], sharey=ax_variables)
    yticks = _np.arange(n_rows)

    # axes without ticks (first/textual ones)
    for ax in [ax_variables, ax_groups, ax_n, ax_ci, ax_pval]:
        ax.set_yticks([])
        ax.set_xticks([])
        ax.tick_params(left=False, right=False, top=False, bottom=False)
        ax.spines[:].set_visible(False)

    # variable column
    ax_variables.set_ylim(-0.5, n_rows - 0.5)
    ax_variables.set_xlim(0, 1)
    ax_variables.invert_yaxis()
    ax_variables.set_title("Variables", fontsize=fontsize, pad=12, fontfamily=fontfamily, fontweight="bold", loc="left")
    for y, val in enumerate(df[variable_col]):
        ax_variables.text(0, y, str(val), fontfamily=fontfamily, fontsize=fontsize, va="center", ha="left")

    # groups column
    ax_groups.set_ylim(ax_variables.get_ylim())
    ax_groups.set_xlim(0, 1)
    ax_groups.set_title("Groups", fontsize=fontsize, pad=12, fontfamily=fontfamily, fontweight="bold", loc="left")
    for y, val in enumerate(df[group_col]):
        ax_groups.text(0, y, str(val), fontfamily=fontfamily, fontsize=fontsize, va="center", ha="left")

    # n column
    ax_n.set_ylim(ax_variables.get_ylim())
    ax_n.set_xlim(0, 1)
    ax_n.set_title("n", fontsize=fontsize, pad=12, fontfamily=fontfamily, fontweight="bold", loc="center")
    for y, val in enumerate(n_texts):
        ax_n.text(0.5, y, val, fontfamily=fontfamily, fontsize=fontsize, va="center", ha="center")

    # CI column
    ax_ci.set_ylim(ax_variables.get_ylim())
    ax_ci.set_xlim(0, 1)
    ax_ci.set_title("Est (CI)", fontsize=fontsize, pad=12, fontfamily=fontfamily, fontweight="bold",loc="left")
    for y, val in enumerate(ci_texts):
        ax_ci.text(0, y, val, fontfamily=fontfamily, fontsize=fontsize, va="center", ha="left")

    # pvalue column
    ax_pval.set_ylim(ax_variables.get_ylim())
    ax_pval.set_xlim(0, 1)
    ax_pval.set_title("p", fontsize=fontsize, pad=12, fontfamily=fontfamily, fontweight="bold", loc="right")
    for y, val in enumerate(pval_texts):
        ax_pval.text(1, y, val, fontfamily=fontfamily, fontsize=fontsize, va="center", ha="right")

    # forestplot column
    ax_forest.set_ylim(ax_variables.get_ylim())
    ax_forest.set_yticks(yticks)
    ax_forest.set_yticklabels([])
    if forest_title != "":
        ax_forest.set_title(forest_title, fontsize=fontsize, pad=12, fontfamily=fontfamily, fontweight="bold", loc="center")
    for y, (_, row) in enumerate(df.iterrows()):
        est = row[est_col]
        ll = row[ll_col]
        hl = row[hl_col]
        # confidence interval line plotting
        # print CI only if CI is actually available (and estimate is not 0)
        if isinstance(ll, float) and isinstance(hl, float):
            ax_forest.plot([ll, hl], [y, y], color=ci_color, lw=2.3, zorder=1)
        if est > 0.000001 and abs(est-1) > 0.000001:  # do not print 0 or 1
            ax_forest.plot(est, y, "o", color=dot_color, markersize=9, zorder=2)
        ax_forest.set_xlim(*forest_xlim)
        if log_scale:
            ax_forest.set_xscale('log')  # base 10
            # ax_forest.set_xscale('log', base=2)

    ax_forest.axvline(refline, color=refline_color, ls=refline_style, lw=2, zorder=0)
    ax_forest.tick_params(axis="y", left=False, right=False, labelleft=False)
    ax_forest.tick_params(axis="x", labelsize=fontsize - 2)
    ax_forest.set_xlabel(xlabel, fontsize=fontsize + 1, fontfamily=fontfamily)

    ax_forest.spines["left"].set_visible(False)
    ax_forest.spines["right"].set_visible(False)
    ax_forest.grid(axis="x", linestyle=":", alpha=0.7)
    fig.align_ylabels([ax_variables, ax_groups, ax_n, ax_ci, ax_pval, ax_forest])
    return fig



"""Handy confidence intervals"""

import pandas as _pd
from pylbmisc.r import match_arg as _match_arg
from scipy import stats as _stats


def ci_prop(x, n=None, nas=0, confidence_level=0.95,
            method="exact",
            alternative="two-sided"
            ):
    """Confidence interval for single proportion

    It's 95% two-sided exact Clopper-Pearson by default.

    Parameters
    ----------
    x: int or pd.Categorical with two categories
        variable or count of successes
    n: int
        if x is integer, number of trials
    nas: int
        if x is integer, number of missing values out of n
    confidence_level: float
        confidence level for confidence interval
    method: str
        one between "exact" (default Clopper-Pearson), "wilson" (Wilson without
        continuity correction) "ccwilson" (Wilson with continuity correction)
        or any of their abbreviations
    alternative: str
        one between "two-sided", "greater", "less" or their abbreviations

    Examples
    --------
    >>> # single variable
    >>> ci_prop(df.adesione_intervento_proposto)
    >>> ci_prop(27, 27+21, 34)

    >>> # several variables stratified by time
    >>> main_vars = ['adesione_intervento_posto_di_lavoro', 'proposta_posto_di_lavoro',
    ...              'adesione_intervento_proposto', 'follow_up', 'soddisfazione',
    ...              'rtw_continuazione_del_lavoro_alla_fine_del_follow_up']
    >>>
    >>> res = {}
    >>> for var in main_vars:
    ...     res[var] = (
    ...           df.loc[:, var].groupby(df.time).apply(lambda x: ci_prop(x=x))
    ...                         .reset_index().drop(columns = ["level_1"])
    ... )

    """
    method = _match_arg(method, ["exact", "wilson", "ccwilson"])
    alternative = _match_arg(alternative, ["two-sided", "greater", "less"])
    # move to scipy.stats._result_classes.BinomTestResult.proportion_ci naming
    # wilsoncc
    if method == "ccwilson":
        method = "wilsoncc"

    if n is None:
        # caso in cui passo una serie: guarda le labels (la seconda,
        # tipicamente "Yes" sarà usata per il numeratore)
        groups = x.cat.categories
        first_group = groups[0]
        second_group = groups[1]
        na = x.isna().sum()
        first_group_n = x.isin([first_group]).sum()
        second_group_n = x.isin([second_group]).sum()
    else:
        # caso in cui passo delle conte già fatte
        first_group = "unsuccesses"
        second_group = "successes"
        na = nas
        first_group_n = n - x - na
        second_group_n = x

    try:
        binom_test = _stats.binomtest(
            k=second_group_n,
            n=first_group_n + second_group_n,
            alternative=alternative
        )
        ci = binom_test.proportion_ci(
            confidence_level=confidence_level,
            method=method
        )
    except Exception:  # fix this someday
        est = _pd.NA
        lower = _pd.NA
        upper = _pd.NA
    else:
        est = binom_test.statistic
        lower = ci.low
        upper = ci.high

    return _pd.DataFrame({
        "n": [na + first_group_n + second_group_n],
        "NA": [na],
        first_group: [first_group_n],
        second_group: [second_group_n],
        "perc": [est * 100],
        "ci_lower": [lower * 100],
        "ci_upper": [upper * 100]
    })



# SPDX-FileCopyrightText: 2023-present Luca Braglia <lbraglia@gmail.com>
#
# SPDX-License-Identifier: MIT

from pylbmisc import datasets, dm, fig, io, iter, rand, r, \
    scorer, stats, surv, utils

# from pylbmisc.__about__ import __version__


__all__ = ["datasets", "dm", "fig", "io", "iter", "rand", "scorer",
           "stats", "surv", "utils", "r"]



"""
Class and utilities for randomization

The main class is List which produces a blocked randomizzation list with
optional stratification by several factors.
"""

from pylbmisc.r import expand_grid as _expand_grid
import numpy as _np
import pandas as _pd
from itertools import permutations as _permutations
from itertools import repeat as _repeat
from pathlib import Path as _Path
from pprint import pp as _pp
from pprint import pformat as _pformat


class List:
    """Stratified/blocked randomization list generation

    Examples
    --------
    >>> # i centri debbono essere il PRIMO criterio di stratificazione
    >>> # altrimenti l'aggiunta di centri in corso dello studio può incasinare
    >>> # la randomizzazione (delle liste gia fatte)
    >>> strata = {"centres": ["ausl_re", "ausl_mo"],
    ...           "agecl": ["<18", "18-65", ">65"]}
    >>> a = lb.rand.List(seed=354, n = 100, strata=strata)
    >>> a.stats()
    >>> a.to_txt() # <- in "/tmp"
    >>> a.to_csv() # <- in "/tmp"
    """

    def __init__(self,
                 seed=None,
                 groups=["Control", "Experimental"],  # example 1:1 ratio with
                 reps=[1, 2, 3],                      # blocks 2, 4 or 6
                 n=100,                               # trial sample size
                 strata={"centres": ["ausl_re"]}):
        if seed is None:
            msg = "Must specify a seed."
            raise ValueError(msg)
        self._rng = _np.random.default_rng(seed=seed)
        self._n = n
        self._strata_df = _expand_grid(strata)
        # sample blocks to be permuted eg [["C", "T"], ["C", "T", "C", "T"],
        # ["C", "T", "C", "T", "C", "T"]]
        blocks = []
        for rep in reps:
            blocks.append(groups * rep)
        # Actual length: eg [2, 4, 6]
        self._blocks_len = [len(b) for b in blocks]
        # Pool of all the permutations to be used to pick one
        pool = []
        for block in blocks:
            # do permutations, use set to remove duplicates and list to sort
            unique_block_perms = list(set(_permutations(block)))
            unique_block_perms.sort()
            pool.append(unique_block_perms)
        self._pool = pool
        # Actually generate and store the randomization list
        randlist = []
        for _, row in self._strata_df.iterrows():
            randlist.append({
                "strata": row,
                "rl": self._generate_strata_list()})
        self._randlist = randlist

    def _generate_strata_list(self):
        actual_n = 0
        block_id = 0
        unit_id = 0
        strata_randlist = []
        while (actual_n < self._n):
            # increase counter of randomized blocks
            block_id += 1
            # select a random len
            block_sel = int(self._rng.integers(len(self._blocks_len)))
            # obtain the actual len
            block_dim = self._blocks_len[block_sel]
            # how many permutations are available for a given dim
            n_avail = len(self._pool[block_sel])
            chosen_block = int(self._rng.integers(n_avail))  # chose one
            sampled_block = self._pool[block_sel][chosen_block]   # which one
            actual_n += block_dim                       # add to sample size
            for b_id, b_dim, trt in zip(_repeat(block_id),  # create the df
                                        _repeat(block_dim),
                                        sampled_block):
                unit_id += 1
                strata_randlist.append({
                    "unit": unit_id,
                    "block": b_id,
                    "block_dim": b_dim,
                    "trt": trt
                })
        return _pd.DataFrame(strata_randlist)

    def stats(self):
        """Print stats of the randomization list"""
        stratas_stats = []  # list of dicts
        for stratalist in self._randlist:
            strata = stratalist["strata"]
            rl = stratalist["rl"]
            strata_string = "_".join(strata.to_list())
            rval = {
                "strata": strata_string,
                "n": rl.shape[0],
                "n_blocks": len(rl.block.unique()),
                "overall_balance": len(rl.trt.value_counts().unique()) == 1,
                "block_balance": _pd.crosstab(rl.block, rl.trt).apply(
                    # all the rows must have the same frequencies
                    axis="columns",  # apply the function by row
                    func=lambda x: len(x.unique()) == 1,  # check row same freq
                    ).all()}  # return True if all check are ok
            stratas_stats.append(rval)
        stratas_stats = _pd.DataFrame(stratas_stats)
        with _pd.option_context("display.max_rows", None,
                                "display.max_columns", None):
            print(stratas_stats)

    def __repr__(self):
        with _pd.option_context("display.max_rows", None,
                                "display.max_columns", None):
            return _pformat(self._randlist, indent=0, sort_dicts=False)

    def to_txt(self,
               fpath: str | _Path = _Path("/tmp/randomization_list.txt")
               ) -> None:
        """Export randomization list to text file

        param:
        fpath: file path
        """
        with _Path(fpath).open("w") as f:
            # print all the rows thanks
            with _pd.option_context("display.max_rows", None,
                                    "display.max_columns", None):
                for stratalist in self._randlist:
                    print("-" * 82, file=f)
                    _pp(stratalist["strata"], stream=f)
                    print("", file=f)
                    _pp(stratalist["rl"], stream=f)
                    print("", file=f)

    def to_csv(self, dpath: str | _Path = "/tmp/"
               ) -> None:
        """Export randomization lists to csv(s) in a directory

        param:
        dpath: the directory path
        """
        dpath = _Path(dpath)
        for stratalist in self._randlist:
            strata = stratalist["strata"]
            strata_string = "_".join(strata.to_list())
            actual_path = dpath / f"{strata_string}.csv"
            stratalist["rl"].to_csv(actual_path, index=False)


if __name__ == "__main__":
    # i centri debbono essere il PRIMO criterio di stratificazione altrimenti
    # l'aggiunta di centri in corso dello studio può incasinare la
    # randomizzazione (delle liste gia fatte)
    strata = {"centres": ["ausl_re", "ausl_mo", "ausl_bo"],
              "agecl": ["lt_18", "18_65", "gt_65"]}
    a = List(seed=354, strata=strata)
    a.stats()
    a.to_txt()  #<- in /tmp
    a.to_csv()  #<- in /tmp



"""Data management utilities for pandas Series/DataFrame"""

import functools as _functools
import inspect as _inspect
import numpy as _np
import pandas as _pd
import pyarrow as _pa
import re as _re
import string as _string
import datetime as _dt

from collections import Counter as _Counter
from functools import singledispatch as _singledispatch
from pathlib import Path as _Path
from pprint import pprint as _pprint


_default_dtype_backend = "pyarrow"

# -------------------------------------------------------------------------
# regular expressions used
# -------------------------------------------------------------------------

# Searching by regex in data
# https://stackoverflow.com/questions/51170763 for general setup
# Mail: https://stackoverflow.com/questions/8022530/ for mail regex
_mail_re = _re.compile(r"[^@]+@[^@]+\.[^@]+")
_fc_re = _re.compile(
    r"[A-Za-z]{6}[0-9]{2}[A-Za-z]{1}[0-9]{2}[A-Za-z]{1}[0-9]{3}[A-Za-z]{1}"
)
_tel_re = _re.compile(r"(.+)?0[0-9]{1,3}[\. /\-]?[0-9]{6,7}")
_mobile_re = _re.compile(r"(.+)?3[0-9]{2}[\. /\-]?[0-9]{6,7}")


_is_mobile_number = _np.vectorize(lambda x: bool(_mobile_re.match(x)))
_is_telephone_number = _np.vectorize(lambda x: bool(_tel_re.match(x)))
_is_email = _np.vectorize(lambda x: bool(_mail_re.match(x)))
_is_fiscal_code = _np.vectorize(lambda x: bool(_fc_re.match(x)))

# cleaning string that have to contain dates
_dates_polish = _re.compile(r"[^/\d-]") # keep only numbers, - and /, and just hope for the best



# -------------------------------------------------------------------------
# Utilities
# -------------------------------------------------------------------------
def table2df(df: _pd.DataFrame) -> _pd.DataFrame:
    """Transform a pd.DataFrame representing a two-way table (es
    crosstable, correlation matrix, p.val matrix) in a
    pd.DataFrame with long format.

    Parameters
    ----------
    df:
       the crosstabulation to be put in long form

    Examples
    --------
    >>> import pandas as pd
    >>> a = pd.Series(["foo", "foo", "foo", "foo", "bar", "bar",
    ...                "bar", "bar", "foo", "foo", "foo"], dtype=object)
    >>> b = pd.Series(["one", "one", "one", "two", "one", "one",
    ...                "one", "two", "two", "two", "one"], dtype=object)
    >>> pd.crosstab(a, b)
    col_0  one  two
    row_0
    bar      3    1
    foo      4    3
    >>> tab = pd.crosstab(a,b)
    >>> table2df(tab)
      row_0 col_0  x
    0   bar   one  3
    1   bar   two  1
    2   foo   one  4
    3   foo   two  3
    """
    if not isinstance(df, _pd.DataFrame):
        msg = "Only dataframes are processed."
        raise Exception(msg)
    x = df.copy()
    x = x.stack().reset_index()
    return x.rename(columns={0: "x"})


def dump_unique_values(dfs: _pd.DataFrame | dict[str, _pd.DataFrame],
                       fpath: str | _Path = "tmp/uniq.txt"):
    """Save unique value of a (dict of) dataframe for inspection and monitor
    during time.

    Parameters
    ----------
    dfs:
        dataframe or dict of dataframes to be dumped
    fpath:
        path where to save the unique values
    """
    if not isinstance(dfs, (_pd.DataFrame, dict)):
        msg = "x deve essere un pd.DataFrame o un dict di pd.DataFrame"
        raise ValueError(msg)
    # normalize single dataframe
    if isinstance(dfs, _pd.DataFrame):
        dfs = {"df": dfs}
        # output
    outfile = _Path(fpath)
    with outfile.open("w") as f:
        for df_lab, df in dfs.items():
            for col in df:
                # Header
                print(f"DataFrame: '{df_lab}', "
                      f"Column: '{col}', "
                      f"Dtype: {df[col].dtype}, "
                      f"Unique values:",
                      file=f)
                # Dati: non sortati perché ci sono problemi se i dati sono metà
                # numerici e meta stringa? bah ci riprovo
                if is_string(df[col]):
                    _pprint(df[col].value_counts().index.to_list(),
                            stream=f,
                            compact=True)
                else:
                    _pprint(df[col].sort_values().unique().tolist(),
                            stream=f,
                            compact=True)
                print(file=f)
            print("-"*80 + "\n\n", file=f)


def names_list(dfs):
    """Dump names to stdout to ease code creation for Coercer

    Parameters
    ----------
    dfs:
        dataframe or dict of dataframes to be dumped
    """
    if isinstance(dfs, dict):
        for k, df in dfs.items():
            print(k, "\n")
            _pprint(df.columns.to_list())
    elif isinstance(dfs, _pd.DataFrame):
        _pprint(dfs.columns.to_list())
    else:
        msg = "Not handled data (only dataframe or dicts of dataframes)"
        raise Exception(msg)


def qcut(x, q, **kwargs) -> _pd.Categorical:
    """An alternative to pd.qcut

    This function produces categorization based on quantiles but without the
    index produced by pd.qcut that can give issues in some routines not

    Parameters
    ----------
    x: 1d ndarray or Series
        as in pd.qcut
    q: int or list-like of float
        as in pd.qcut
    kwargs:
        other parameter passed to pd.qcut (all but labels!)
    """
    full = _pd.qcut(x, q, **kwargs)
    categories = full.cat.categories.astype(str).to_list()
    raw = _pd.qcut(x, q, labels=False, **kwargs)
    recode = {prog: lab for prog, lab in enumerate(categories)}
    mapped = raw.map(recode)
    return to_categorical(mapped, levels=categories)


# -------------------------------------------------------------------------
# Tests
# ------------------------------------------------------------------------

def is_all_missing(x: _pd.Series) -> bool:
    """Check if all the Series is composed of missing values

    Check if all the Series is composed of missing values

    Parameters
    ----------
    x:
        the Series to be checked
    """
    return x.isna().all()


def is_string(x: _pd.Series) -> bool:
    """Check the type of a Series to be composed of string.

    It checks Series (including data with np.nan) differently from
    pandas.api.types.is_string_dtype

    Parameters
    ----------
    x:
        the Series to be checked

    Examples
    --------
    >>> import pandas as pd
    >>> x = pd.Series(["a", pd.NA])
    >>> pd.api.types.is_string_dtype(x)
    False
    >>> is_string(x)
    True
    """
    return x.dtype in ["O", "string[pyarrow]"]


def is_bool(x: _pd.Series) -> bool:
    """Check the type of a Series to be composed of boolean.

    Parameters
    ----------
    x:
        the Series to be checked
    """
    return _pd.api.types.is_bool_dtype(x)


def is_integer(x: _pd.Series) -> bool:
    """Check the type of a Series to be composed of integers.

    Parameters
    ----------
    x:
        the Series to be checked
    """
    return _pd.api.types.is_integer_dtype(x)


def is_numeric(x: _pd.Series) -> bool:
    """Check the type of a Series to be composed of numerics.

    Parameters
    ----------
    x:
        the Series to be checked
    """
    return _pd.api.types.is_numeric_dtype(x)


def is_categorical(x: _pd.Series) -> bool:
    """Check the type of a Series to be composed of string.

    It checks Series (including data with np.nan) differently from
    pandas.api.types.is_string_dtype

    Parameters
    ----------
    x:
        the Series to be checked
    """
    return _pd.api.types.is_categorical_dtype(x)


def is_datetime(x: _pd.Series) -> bool:
    """Check the type of a Series to be composed of datetimes.

    Parameters
    ----------
    x:
        the Series to be checked
    """
    return _pd.api.types.is_datetime64_any_dtype(x)


def is_date(x: _pd.Series) -> bool:
    """Check the type of a Series to be composed of dates.

    Parameters
    ----------
    x:
        the Series to be checked
    """
    # If it's not a datetime it can't be a date
    if not is_datetime(x):
        return False

    # otherwise check if all the times are zero; fix na for the sake of god:
    # don't put in all the comparison based on NA
    zero_time = (x.dt.time == _dt.time(0, 0))
    zero_time = zero_time[~ x.isna()]
    return zero_time.all()





# -------------------------------------------------------------------------
# pii_erase
# ------------------------------------------------------------------------
# Searching by column name
def _columns_match(df_columns, searched):
    """Search in columns names, both exact match and contains match"""
    # coerce single string to list of one string
    if isinstance(searched, str):
        searched = [searched]
    perfect_match = [c in set(searched) for c in df_columns]
    return perfect_match



def _has_emails(x: _pd.Series) -> bool:
    """ Check if a Series has e-mail address

    Parameters
    ----------
    x:
        the Series to be checked

    """
    if is_string(x):
        check = _is_email(x)
        return bool(_np.any(check))
    else:
        return False



def _has_fiscal_codes(x: _pd.Series) -> bool:
    """Check if a Series has fiscal codes

    Parameters
    ----------
    x:
        the Series to be checked
    """
    if is_string(x):
        check = _is_fiscal_code(x)
        return bool(_np.any(check))
    else:
        return False



def _has_telephone_numbers(x: _pd.Series) -> bool:
    """Check if a Series has telephone numbers

    Parameters
    ----------
    x:
        the Series to be checked

    """
    if is_string(x):
        check = _is_telephone_number(x)
        return bool(_np.any(check))
    else:
        return False



def _has_mobile_numbers(x: _pd.Series) -> bool:
    """Check if a Series has mobile numbers

    Parameters
    ----------
    x:
        the Series to be checked
    """
    if is_string(x):
        check = _is_mobile_number(x)
        return bool(_np.any(check))
    else:
        return False


def pii_find(x: _pd.DataFrame) -> list[str]:
    """Find columns with probable piis (Personally Identifiable
    Informations) and return the colnames for further processing.

    Parameters
    ----------
    x:
        The DataFrame to check

    Returns
    -------
    A Python list of variable names with probable PII

    Examples
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({
    ...     "id" : [1,2,3],
    ...     "cognome": ["brazorv", "gigetti", "ginetti"],
    ...     "nome  " : [1,2,3],
    ...     "mail": ["lgasd@asdkj.com", " asòdlk@asd.com", "aaaa"],
    ...     "fc": ["nrgasd12h05h987z", "aaaa", "eee"],
    ...     "num": ["0654-6540123", "aa", "eee"],
    ...     "cel": ["3921231231", "aa", "eee"]
    ...     })
    >>>
    >>> probable_piis = pii_find(df)
    'cognome' matches 'surname'/'cognome'.
    'nome  ' matches 'name'/'nome'.
    mail probably contains emails.
    fc probably contains fiscal codes.
    num probably contains telephone numbers.
    cel probably contains mobile phones numbers.
    >>> probable_piis
    ['cognome', 'nome  ', 'mail', 'fc', 'num', 'cel']
    >>> if probable_piis:
    ...     df.drop(columns=probable_piis)
    ...
       id
    0   1
    1   2
    2   3
    """

    if not isinstance(x, _pd.DataFrame):
        msg = "x must be a pd.DataFrame"
        raise ValueError(msg)

    col = list(x.columns.values)

    # name and surname (looking at columns names)
    col_clean = [c.lower().strip() for c in col]
    surname_match = _columns_match(col_clean, ["cognome", "surname"])
    name_match = _columns_match(col_clean, ["nome", "name"])

    # finding other pii loking at data
    has_mails = [_has_emails(x[c]) for c in col]
    has_fcs = [_has_fiscal_codes(x[c]) for c in col]
    has_telephones = [_has_telephone_numbers(x[c]) for c in col]
    has_mobiles = [_has_mobile_numbers(x[c]) for c in col]

    probable_pii = []
    zipped = zip(
        col,
        surname_match,
        name_match,
        has_mails,
        has_fcs,
        has_telephones,
        has_mobiles,
    )
    for (
        var,
        is_surname,
        is_name,
        has_mail,
        has_fc,
        has_telephone,
        has_mobile,
    ) in zipped:
        pii_sum = (
            is_surname
            + is_name
            + has_mail
            + has_fc
            + has_telephone
            + has_mobile
        )
        if pii_sum:
            probable_pii.append(var)
            if is_surname:
                print(f"'{var}' matches 'surname'/'cognome'.")
            if is_name:
                print(f"'{var}' matches 'name'/'nome'.")
            if has_mail:
                print(f"{var} probably contains emails.")
            if has_fc:
                print(f"{var} probably contains fiscal codes.")
            if has_telephone:
                print(f"{var} probably contains telephone numbers.")
            if has_mobile:
                print(f"{var} probably contains mobile phones numbers.")
    return probable_pii


# -------------------------------------------------------------------------
# preprocessing varnames (from shitty excel): fix_columns
# -------------------------------------------------------------------------


def _compose(f, g):
    return lambda x: f(g(x))


def _replace_unwanted_chars(s):
    keep = _string.ascii_lowercase + _string.digits
    for ch in s:
        if ch not in keep:
            s = s.replace(ch, "_")
    return s


def _replace_accents(s):
    tmp = _re.sub("à", "a", s)
    tmp = _re.sub("è", "e", tmp)
    tmp = _re.sub("é", "e", tmp)
    tmp = _re.sub("ì", "i", tmp)
    tmp = _re.sub("ù", "u", tmp)
    return tmp


def _remove_duplicated_underscore(s):
    return _re.sub("_+", "_", s)


def _remove_external_underscore(s):
    return _re.sub("^_", "", _re.sub("_$", "", s))


def _add_x_if_first_is_digit(s):
    if s.startswith(tuple(_string.digits)):
        return "x" + s
    else:
        return s


def _fix_varnames_worker(vnames: list[str],
                         make_unique: bool) -> list[str]:
    funcs = [
        lambda s: str(s),
        lambda s: s.lower().strip(),
        lambda s: _replace_accents(s),
        lambda s: _replace_unwanted_chars(s),
        lambda s: _remove_duplicated_underscore(s),
        lambda s: _remove_external_underscore(s),
        lambda s: _add_x_if_first_is_digit(s),
    ]
    # sotto serve dato che compose applica prima la funzione a destra
    # e poi quella a sx (come in math) mentre vogliamo tenere l'ordine
    # di sopra
    funcs.reverse()
    worker = _functools.reduce(_compose, funcs)
    mod = [worker(s) for s in vnames]
    # handle duplicated names by adding numeric postfix
    has_duplicates = len(mod) != len(set(mod))
    if has_duplicates and make_unique:
        seen = {}
        uniq = []
        for v in mod:
            if v not in seen:
                seen[v] = 0
                uniq.append(v)
            else:
                seen[v] += 1
                uniq.append(f"{v}_{seen[v]}")
    else:
        uniq = mod
    # se ci sono doppi nei nomi di partenza meglio evitare i dict se no i
    # doppi vengono considerati solo una volta
    # rename_dict = {}
    # for o, m in zip(original, uniq_mod):
    #     rename_dict.update({o: m})
    # if it was a string that was passed, return a string, not a list
    return uniq


def fix_varnames(x: str | list[str] | _pd.Series | _pd.DataFrame | dict[str, _pd.DataFrame],
                 return_tfd: bool = False,
                 make_unique: bool = True,
                 ):
    """The good-old R preprocess_varnames, working with strings,
    lists, pd.DataFrame or dicts of pd.DataFrames.

    Parameters
    ----------
    x:
        string, list of strigs/varnames, series, dataframe or dict of dataframes
    return_tfd:
        bool, return a dict of to/from descriptions
    make_unique:
        bool, assure returned varnames are unique by adding a progressive number as postfix

    Examples
    --------
    >>> fix_varnames("  asd 98n2 3")
    'asd_98n2_3'
    >>> fix_varnames([" 98n2 3", " L< KIAFJ8 0_________"])
    ['x98n2_3', 'l_kiafj8_0']
    >>> fix_varnames(["àsd", "foo0", "asd"])
    ['asd', 'foo0', 'asd_1']

    """

    if isinstance(x, str):
        x = [x]

    if isinstance(x, list):
        from_name = x
        to_name = _fix_varnames_worker(from_name, make_unique=make_unique)
        to_name = to_name if len(to_name) > 1 else to_name[0]
        if return_tfd:
            tf = {t: f for t, f in zip(to_name, from_name)}
            return to_name, tf
        else:
            return to_name
    elif isinstance(x, _pd.Series):
        from_name = x.to_list()
        to_name = _fix_varnames_worker(from_name, make_unique=make_unique)
        rval = _pd.Series(to_name, index=x.index).astype(_pd.ArrowDtype(_pa.string()))
        if return_tfd:
            tf = {t: f for t, f in zip(to_name, from_name)}
            return rval, tf
        else:
            return rval
    elif isinstance(x, _pd.DataFrame):
        from_name = list(x.columns.values)
        to_name = _fix_varnames_worker(from_name, make_unique=make_unique)
        df = x.copy()
        df.columns = to_name
        if return_tfd:
            tf = {t: f for t, f in zip(to_name, from_name)}
            return df, tf
        else:
            return df
    elif isinstance(x, dict):
        dfs = {}
        tfs = {}
        for k, v in x.items():
            from_name = list(v.columns.values)
            to_name = _fix_varnames_worker(from_name, make_unique=make_unique)
            df = v.copy()
            df.columns = to_name
            tf = {t: f for t, f in zip(to_name, from_name)}
            dfs[k] = df
            tfs[k] = tf
        if return_tfd:
            return dfs, tfs
        else:
            return dfs
    else:
        msg = "Only str and list are allowed, see sanitize_varnames for DataFrame or dict of DataFrames."
        raise ValueError(msg)




# -------------------------------------------------------------------------
# Coercion stuff below
# -------------------------------------------------------------------------
# decorators
def _verboser(f):
    def transformer(x: _pd.Series):
        coerced = f(x)
        report = _pd.DataFrame({"original": x, "coerced": coerced})
        new_na = (~_pd.isna(x)) & (_pd.isna(coerced))
        if _np.any(new_na):
            print("Please check if the following coercions are ok:")
            print(report[new_na])
        return coerced
    return transformer


# --------------- coercion workers ----------------------------------------

def identity(x):
    """The identity function to coerce nothing

    Parameters
    ----------
    x: Any
      any data

    Returns
    -------
    the object x unchanged
    """
    return x


def to_bool(x=None) -> _pd.Series:
    """Coerce to a boolean pd.Series

    Parameters
    ----------
    x: Series or something coercible to
        data to be coerced

    Examples
    --------
    >>> import pandas as pd
    >>> to_bool([1,0,1,0])
    0     True
    1    False
    2     True
    3    False
    dtype: bool[pyarrow]
    >>> to_bool(pd.Series([1,0.,1,0.]))
    0     True
    1    False
    2     True
    3    False
    dtype: bool[pyarrow]
    >>> to_bool([1,0,1,0, pd.NA])
    0     True
    1    False
    2     True
    3    False
    4     <NA>
    dtype: bool[pyarrow]
    """
    if x is None:
        msg = "x must be a Series or something coercible to, not None."
        raise ValueError(msg)
    if not isinstance(x, _pd.Series):
        x = _pd.Series(x)
    nas = _pd.isna(x)
    rval = x.astype("boolean[pyarrow]" if _default_dtype_backend == "pyarrow" else "boolean")
    rval[nas] = _pd.NA
    return rval


def _replace_comma(x: _pd.Series):
    if is_string(x):
        # nas = x.isin(["", _pd.NA, _np.nan])
        nas = (x.isna()) | (x == "")
        rval = x.astype("str").str.replace(",", ".")
        rval[nas] = _pd.NA
        return rval
    else:
        return x


# actually to_integer is used only by tteep to ensure, otherwise to_numeric
# with pyarrow backend should handle both integers and floats gracefully and
# there's no need for another function
def to_integer(x=None) -> _pd.Series:
    """Coerce a pd.Series to integer (if possible)

    Parameters
    ----------
    x: Series or something coercible to
        data to be coerced

    Examples
    --------
    >>> import numpy as np
    >>> to_integer([1., 2., 3., 4., 5., 6., np.nan])
    0       1
    1       2
    2       3
    3       4
    4       5
    5       6
    6    <NA>
    dtype: Int64
    >>> to_integer(["2001", "2011", "1999", ""])
    0    2001
    1    2011
    2    1999
    3    <NA>
    dtype: Int64
    >>> # to_integer(["1.1", "1,99999", "foobar"]) # fails because of 1.99
    """
    if x is None:
        msg = "x must be a Series or something coercible to, not None."
        raise ValueError(msg)
    if not isinstance(x, _pd.Series):
        x = _pd.Series(x)
    s = _replace_comma(x)
    # return _np.floor(_pd.to_numeric(s, errors='coerce')).astype('Int64')
    # mi fido piu di quella di sotto anche se fallisce con i numeri con virgola
    # (che ci può stare, per i quale bisogna usare np.floor)
    return _pd.to_numeric(
        s,
        # dtype_backend = _default_dtype_backend,
        errors="coerce").astype(
            "Int64"
            # "int64[pyarrow]" if _default_dtype_backend == 'pyarrow' else "Int64"
            # _pd.ArrowDtype(_pa.int64()) if _default_dtype_backend == 'pyarrow' else "Int64"
        )


def to_numeric(x=None) -> _pd.Series:
    """Coerce a pd.Series using pd.to_numeric

    Parameters
    ----------
    x: Series or something coercible to
        data to be coerced

    Examples
    --------
    >>> to_numeric([1, 2, 3])
    0    1
    1    2
    2    3
    dtype: int64[pyarrow]
    >>> to_numeric([1., 2., 3., 4., 5., 6.])
    0    1.0
    1    2.0
    2    3.0
    3    4.0
    4    5.0
    5    6.0
    dtype: double[pyarrow]
    >>> to_numeric(["2001", "2011", "1999"])
    0    2001
    1    2011
    2    1999
    dtype: int64[pyarrow]
    >>> to_numeric(["1.1", "2,1", "asd", ""])
    0     1.1
    1     2.1
    2    <NA>
    3    <NA>
    dtype: double[pyarrow]
    """
    if x is None:
        msg = "x must be a Series or something coercible to, not None."
        raise ValueError(msg)
    if not isinstance(x, _pd.Series):
        x = _pd.Series(x)
    s = _replace_comma(x)
    return _pd.to_numeric(s, errors="coerce",
                          dtype_backend=_default_dtype_backend)


def to_datetime(x=None, *args, **kwargs) -> _pd.Series:
    """Coerce to a datetime a pd.Series

    Parameters
    ----------
    x: Series or something coercible to
        data to be coerced
    args:
        other arguments passed to pd.to_datetime
    kwargs:
        other arguments passed to pd.to_datetime
    
    Examples
    --------
    >>> to_datetime(['2025-03-01 10:02:22.756611'] * 2)
    0   2025-03-01 10:02:22.756611
    1   2025-03-01 10:02:22.756611
    dtype: datetime64[ns]
    """
    if x is None:
        msg = "x must be a Series or something coercible to, not None."
        raise ValueError(msg)
    if not isinstance(x, _pd.Series):
        x = _pd.Series(x)
    return _pd.to_datetime(x, errors="coerce", *args, **kwargs)


def to_date(x=None, *args, **kwargs) -> _pd.Series:
    """Coerce to a date a pd.Series

    Parameters
    ----------
    x: Series or something coercible to
        data to be coerced
    args:
        other arguments passed to pd.to_datetime
    kwargs:
        other arguments passed to pd.to_datetime

    Examples
    --------
    >>> to_date(["2020-01-02", "2021-01-01", "2022-01-02"] * 2)
    0   2020-01-02
    1   2021-01-01
    2   2022-01-02
    3   2020-01-02
    4   2021-01-01
    5   2022-01-02
    dtype: datetime64[ns]
    """
    if x is None:
        msg = "x must be a Series or something coercible to, not None."
        raise ValueError(msg)
    if not isinstance(x, _pd.Series):
        x = _pd.Series(x)
    return to_datetime(x, *args, **kwargs).dt.floor("D")


def _extract_dates_worker(x):
    if isinstance(x, str): # handle missing values (sono float)
        polished = _dates_polish.sub("", x)
        return _pd.to_datetime(polished)
    else:
        # return _np.nan
        return _pd.NA


def extract_dates(x=None) -> _pd.Series:
    """Try to extract dates from shitty strings and convert them to proper

    Parameters
    ----------
    x: Series or something coercible to
        data to be coerced

    Examples
    --------
    >>> extract_dates(["2020-01-02", "01/01/1956", "asdasd 12-01-02"] * 2)
    0   2020-01-02
    1   1956-01-01
    2   2002-12-01
    3   2020-01-02
    4   1956-01-01
    5   2002-12-01
    dtype: datetime64[ns]
    """
    if x is None:
        msg = "x must be a Series or something coercible to, not None."
        raise ValueError(msg)
    if not isinstance(x, _pd.Series):
        x = _pd.Series(x)
    return x.apply(_extract_dates_worker)


def to_categorical(x=None,
                   levels=None,
                   labels=None,
                   ordered: bool = False) -> _pd.Categorical:
    """Coerce to categorical a pd.Series with R's factor API

    Parameters
    ----------
    x: Series or something coercible to
        data to be coerced
    levels: list
        labels to be considered as valid groups, if missing modalities by
        decreasing frequencies will be considered
    labels: list
        labels to be applied to the levels (if missing, will be replaced by
        levels)
    ordered: bool
        make an ordered categorical?

    Examples
    --------
    >>> import numpy as np
    >>> to_categorical([1, 2, 1, 2, 3, np.nan])
    [1.0, 2.0, 1.0, 2.0, 3.0, NaN]
    Categories (3, float64): [1.0, 2.0, 3.0]
    >>> to_categorical(["AA", "BB", "asd", "aa", "", np.nan],
    ...                levels=["AA", "BB"])
    ['AA', 'BB', NaN, NaN, NaN, NaN]
    Categories (2, object): ['AA', 'BB']
    >>> to_categorical(["AA", "BB", "asd", "aa", "", np.nan],
    ...                levels=["AA", "BB"], labels = ["x", "y"])
    ['x', 'y', NaN, NaN, NaN, NaN]
    Categories (2, object): ['x', 'y']
    """
    if x is None:
        msg = "x must be a Series or something coercible to, not None."
        raise ValueError(msg)
    if not isinstance(x, _pd.Series):
        x = _pd.Series(x)
    if levels is None:
        # take levels as from frequencies
        levels = x.value_counts(sort=True, ascending=False).index.to_list()
    if labels is None:
        labels = levels

    if len(levels) != len(labels):
        raise ValueError("levels and labels must have the "
                         "same number of elements")

    levlabs_mapping = {lev: lab for lev, lab in zip(levels, labels)}
    recoded = x.map(levlabs_mapping)
    # ensure labels are unique, https://stackoverflow.com/questions/480214
    unique_labels = list(dict.fromkeys(labels))
    return _pd.Categorical(recoded, categories=unique_labels, ordered=ordered)


def mc(levels=None,
       labels=None,
       ordered: bool = False):
    """Function factory to make categorical variables

    Parameters
    ----------
    levels:
        levels as in to_categorical
    labels:
        labels as in to_categorical
    ordered: bool
        labels as in to_categorical

    Returns
    -------
    a function which coerce to a categorical with specified categories/ordered

    Examples
    --------
    >>> import pylbmisc as lb
    >>> scuola = lb.dm.mc(['Scuola Media Inferiore', 'Scuola Media Superiore',
    ...                    'Laurea', 'Altro'])
    >>> coercer = {
    ...    scuola: ["titolo_istruzione"],
    ... }
    >>>
    """
    def f(x):
        return to_categorical(x=x,
                              levels=levels,
                              labels=labels,
                              ordered=ordered)
    return f


def to_noyes(x=None) -> _pd.Categorical:
    """Coerce to no/yes a string pd.Series

    Parameters
    ----------
    x: Series or something coercible to
        data to be coerced

    Examples
    --------
    >>> import numpy as np
    >>> to_noyes(["","yes","no","boh", "si", np.nan])
    [NaN, 'yes', 'no', NaN, 'yes', NaN]
    Categories (2, object): ['no', 'yes']
    >>> to_noyes([1,0,0, np.nan])
    ['yes', 'no', 'no', NaN]
    Categories (2, object): ['no', 'yes']
    >>> to_noyes([1.,0.,0., np.nan])
    ['yes', 'no', 'no', NaN]
    Categories (2, object): ['no', 'yes']
    """
    if x is None:
        msg = "x must be a Series or something coercible to, not None."
        raise ValueError(msg)
    if not isinstance(x, _pd.Series):
        x = _pd.Series(x)
    if is_string(x):
        # take only the first character and map to n/y
        tmp = x.str.strip().str.lower().astype(str).str[0]
        tmp[tmp == "s"] = "y"
        tmp = tmp.replace({"0": "n", "1": "y"})  # 0/1 for strings
    else:
        # try to convert to boolean and map to n/y
        tmp = to_bool(x).map({False: "n", True: "y"})

    return to_categorical(tmp.map({"n": "no", "y": "yes"}),
                          levels=["no", "yes"])


def to_sex(x=None) -> _pd.Categorical:
    """Coerce to male/female a pd.Series of strings (Mm/Ff)

    Parameters
    ----------
    x: Series or something coercible to
        data to be coerced

    Examples
    --------
    >>> import numpy as np
    >>> to_sex(["","m","f"," m", "Fm", np.nan])
    [NaN, 'male', 'female', 'male', 'female', NaN]
    Categories (2, object): ['male', 'female']
    """
    if x is None:
        msg = "x must be a Series or something coercible to, not None."
        raise ValueError(msg)
    if not isinstance(x, _pd.Series):
        x = _pd.Series(x)
    if not is_string(x):
        msg = "to_sex only for strings vectors"
        raise Exception(msg)

    # take the first letter (Mm/Ff)
    tmp = x.str.strip().str.lower().astype(str).str[0]
    tmp = tmp.map({"m": "male", "f": "female"})
    return to_categorical(tmp, levels=["male", "female"])


def to_recist(x=None) -> _pd.Categorical:
    """Coerce to recist categories a pd.Series of strings

    Parameters
    ----------
    x: Series or something coercible to
        data to be coerced

    Examples
    --------
    >>> import numpy as np
    >>> to_recist(["RC", "PD", "SD", "PR", "RP", "boh", np.nan])
    ['CR', 'PD', 'SD', 'PR', 'PR', NaN, NaN]
    Categories (4, object): ['CR', 'PR', 'SD', 'PD']
    """
    if x is None:
        msg = "x must be a Series or something coercible to, not None."
        raise ValueError(msg)
    if not isinstance(x, _pd.Series):
        x = _pd.Series(x)
    if not is_string(x):
        msg = "to_recist only for strings vectors"
        raise Exception(msg)

    # rm spaces and uppercase and take the first two letters
    tmp = x.str.strip().str.upper().str[:2]
    # uniform italian to english
    ita2eng = {"RC": "CR", "RP": "PR"}
    return to_categorical(tmp.replace(ita2eng),
                          levels=["CR", "PR", "SD", "PD"])


def to_other_specify(x=None) -> _pd.Categorical:
    """Try to polish a bit a free-text variable and create a categorical one

    Parameters
    ----------
    x: Series or something coercible to
        data to be coerced

    Examples
    --------
    >>> to_other_specify(["asd", "asd", "", "prova", "ciao", 3]+ ["bar"]*4)
    ['asd', 'asd', NaN, 'prova', 'ciao', '3', 'bar', 'bar', 'bar', 'bar']
    Categories (5, object): ['bar', 'asd', 'prova', 'ciao', '3']
    """
    if x is None:
        msg = "x must be a Series or something coercible to, not None."
        raise ValueError(msg)
    if not isinstance(x, _pd.Series):
        x = _pd.Series(x)

    nas = (x.isna()) | (x == "")
    tmp = x.copy().astype("str").str.strip().str.lower()
    tmp[nas] = _pd.NA
    # categs ordered by decreasing counts
    categs = list(tmp.value_counts().index)
    return to_categorical(tmp, levels=categs)


def to_string(x=None) -> _pd.Series:
    """Coerce the pd.Series passed to a string Series

    Parameters
    ----------
    x: Series or something coercible to
        data to be coerced

    Examples
    --------
    >>> to_string([1,2,3])
    0    1
    1    2
    2    3
    dtype: object

    """
    if x is None:
        msg = "x must be a Series or something coercible to, not None."
        raise ValueError(msg)
    if not isinstance(x, _pd.Series):
        x = _pd.Series(x)

    nas = x.isna()
    rval = x.astype(_pd.ArrowDtype(_pa.string()))
    rval[nas] = _pd.NA
    return rval


# --------------- coercion class ----------------------------------------
class Coercer:
    """Coercer of a pandas DataFrame.

    Given directives as a dict (variable name as key, function/coercer
    as value) it applies all the function on a copy of the DataFrame
    and return it.  If verbose print a report of the introduced
    missing values with the coercion for check

    Parameters
    ----------
    df:
        The DataFrame to be coerced
    fv:
        function-variable dict: key can be a function or a string containing
        name of the function, variables is a list of strings with name
        of variables to apply the function to
    verbose:
        be verbose about operations applied

    Examples
    --------
    >>> import pylbmisc as lb
    >>> import pandas as pd
    >>> import numpy as np
    >>>
    >>> # ------------------------------------------------
    >>> # Function as key
    >>> # ------------------------------------------------
    >>> raw = pd.DataFrame({
    ...     "idx" :  [1., 2., 3., 4., 5., 6., "2,0", "", np.nan],
    ...     "sex" :  ["m", "maschio", "f", "female", "m", "M", "", "a", np.nan],
    ...     "date":  ["2020-01-02", "2021-01-01", "2022-01-02"] * 2  + [np.nan, "", "a"],
    ...     "state": ["Ohio", "Ohio", "Ohio", "Nevada", "Nevada", "Nevada"] + [np.nan, "", "a"],
    ...     "ohio" : [1, 1, 1, 0, 0, 0] + [np.nan] * 3 ,
    ...     "year":  [str(y) for y in [2000, 2001, 2002, 2001, 2002, 2003]] + [np.nan, "", "a"],
    ...     "pop":   [str(p) for p in [1.5, 1.7, 3.6, np.nan, 2.9]]  + ["3,2", np.nan, "", "a"],
    ...     "recist" : ["", "pd", "sd", "pr", "rc", "cr"] + [np.nan, "", "a"],
    ...     "other" : ["b"]*3 + ["a"]*2 + ["c"] + [np.nan, "", "a"]
    ... })
    >>> directives = {
    ...     lb.dm.to_integer: ["idx", "year"],
    ...     lb.dm.to_sex : ["sex"],
    ...     lb.dm.to_date: ["date"],
    ...     lb.dm.to_categorical: ["state"],
    ...     lb.dm.to_bool: ["ohio"],
    ...     lb.dm.to_numeric: ["pop"],
    ...     lb.dm.to_recist: ["recist"],
    ...     lb.dm.to_other_specify: ["other"]
    ... }
    >>>
    >>> cleaned1 = lb.dm.Coercer(raw, fv = directives, verbose = False).coerce()
    >>> raw
       idx      sex        date   state  ohio  year  pop recist other
    0  1.0        m  2020-01-02    Ohio   1.0  2000  1.5            b
    1  2.0  maschio  2021-01-01    Ohio   1.0  2001  1.7     pd     b
    2  3.0        f  2022-01-02    Ohio   1.0  2002  3.6     sd     b
    3  4.0   female  2020-01-02  Nevada   0.0  2001  nan     pr     a
    4  5.0        m  2021-01-01  Nevada   0.0  2002  2.9     rc     a
    5  6.0        M  2022-01-02  Nevada   0.0  2003  3,2     cr     c
    6  2,0                  NaN     NaN   NaN   NaN  NaN    NaN   NaN
    7             a                       NaN
    8  NaN      NaN           a       a   NaN     a    a      a     a
    >>> cleaned1
        idx     sex       date   state   ohio  year   pop recist other
    0     1    male 2020-01-02    Ohio   True  2000   1.5    NaN     b
    1     2    male 2021-01-01    Ohio   True  2001   1.7     PD     b
    2     3  female 2022-01-02    Ohio   True  2002   3.6     SD     b
    3     4  female 2020-01-02  Nevada  False  2001  <NA>     PR     a
    4     5    male 2021-01-01  Nevada  False  2002   2.9     CR     a
    5     6    male 2022-01-02  Nevada  False  2003   3.2     CR     c
    6     2     NaN        NaT     NaN   <NA>  <NA>  <NA>    NaN   NaN
    7  <NA>     NaN        NaT     NaN   <NA>  <NA>  <NA>    NaN   NaN
    8  <NA>     NaN        NaT       a   <NA>  <NA>  <NA>    NaN     a
    >>>
    >>> # ------------------------------------------------
    >>> # It's possilbe to specify string as key as well
    >>> # ------------------------------------------------
    >>>
    >>> directives2 = {
    ...     "lb.dm.to_categorical": ["state"],
    ...     "lb.dm.to_date": ["date"],
    ...     "lb.dm.to_integer": ["idx", "year"],
    ...     "lb.dm.to_noyes": ["ohio"],
    ...     "lb.dm.to_numeric": ["pop"],
    ...     "lb.dm.to_other_specify": ["other"],
    ...     "lb.dm.to_recist": ["recist"],
    ...     "lb.dm.to_sex" : ["sex"]
    ... }
    >>>
    >>> # same results after ...
    """

    def __init__(
        self,
        df: _pd.DataFrame,
        fv: dict,
        verbose: bool = True,
    ):
        self._df = df
        self._verbose = verbose
        if fv is None:
            msg = "fv can't be None"
            raise ValueError(msg)
        else:
            # Experimental below
            parent_frame = _inspect.currentframe().f_back
            # put it in the vf_dict format evaluating the f in
            # parent frame variable dict
            reversed = {}
            for f, vars in fv.items():
                # if f is a string change it to function taking from the enclosing
                # environment
                f = (
                    eval(f, parent_frame.f_locals, parent_frame.f_globals)
                    if isinstance(f, str)
                    else f
                )
                for v in vars:
                    reversed.update({v: f})
            self._directives = reversed

    def coerce(self, keep_coerced_only: bool = False) -> _pd.DataFrame:
        """Method to apply programmed coercions

        Parameters
        ----------
        keep_coerced_only:
            if True keep only variables in fv dictionary, after coercion
        """
        # do not modify the input data
        df = self._df.copy()
        # keep order of the input variables
        varorder = df.columns.to_list()
        directives = self._directives
        # make verbose all the functions by decorating them
        if self._verbose:
            for var in directives.keys():
                directives[var] = _verboser(directives[var])
        # apply the coercers, but first modify the pd printing options temporarily to
        # handle long reporting of changes
        old_nrows = _pd.get_option("display.max_rows")
        _pd.set_option("display.max_rows", None)
        for var, fun in directives.items():
            if var not in df.columns:
                msg = f"{var} not in df.columns, aborting."
                raise ValueError(msg)
            if self._verbose:
                print(f"Processing {var}.")
            df[var] = fun(df[var])
        _pd.set_option("display.max_rows", old_nrows)
        # return results
        if keep_coerced_only:
            keep = list(directives.keys())
            kept_in_order = [v for v in varorder if v in keep]
            return df[kept_in_order]
        else:
            return df[varorder]


@_singledispatch
def group_prog_id(x):
    """Count the number of times each id was already seen

    Parameters
    ----------
    x: list or np.array or pd.Dataframe
        group indicator

    Examples
    --------
    >>> # no missings
    >>> group_prog_id(3)
    >>> group_prog_id([0, 1, 1, 1, 2, 0, 2, 0])
    >>> group_prog_id(np.array([0, 1, 1, 1, 2, 0, 2, 0]))
    >>> group_prog_id(pd.Series([0, 1, 1, 1, 2, 0, 2, 0]))
    >>> # with missing
    >>> group_prog_id(np.array([1, 1, 0, np.nan, np.nan]))
    >>> group_prog_id(pd.Series([1, 1, 0, pd.NA, pd.NA]))
    """
    raise NotImplementedError(f"{type(x)} is not handled.")


def _group_prog_id_worker(x):
    seen = _Counter()
    rval = []
    for elem in x:
        seen[elem] += 1
        rval.append(seen[elem])
    return rval


@group_prog_id.register
def _(x: list):
    return _group_prog_id_worker(x)


@group_prog_id.register
def _(x: _np.ndarray):
    res = _np.array(_group_prog_id_worker(x))
    return _np.where(_np.isnan(x), _np.nan, res)


@group_prog_id.register
def _(x: _pd.Series):
    rval = _pd.Series(_group_prog_id_worker(x))
    rval[_pd.isna(x)] = _pd.NA
    return to_integer(rval)






# Validation stuff
# ----------------
# if checks are successful if function above are true, than the lambda below
# make false to be true and thus is
class Check:
    """Class representing a single check on a dataframe

    Parameters
    ----------
    fn: function
        function applied to dataframe (must return a pandas series), and a True
        value is considered ok
    errmsg: str
        error message reported for a failed (False) check 
    display_vars: list of variable names
        if provided print the value of variables which are making the check fail
    check_id: str
        a string identifying the check
    """
    def __init__(self, fn, errmsg, display_vars=None, check_id=""):
        if isinstance(display_vars, str):
            display_vars = [display_vars]
        self.fn = fn
        self.errmsg = errmsg
        self.display_vars = display_vars
        self.check_id = check_id


class Validator:
    """Simple rows-checks-first validator of a pandas DataFrame

    Inspired by pandera, given some a list of Checks (ensure) returns all the
    cases which do not fullfill checks, for each one provided

    Parameters
    ----------
    ensure: list of lb.dm.Check
        list containing checks which are ok if True
    df: dataframe
        dataframe to be evaluated
    idcols: str|list
        string or list of columns name of the dataframe which are used to identify
        (failing checks) ros

    Examples
    --------
    # from src/tests/python/validator_tests.py
    # 
    # import pandas as pd
    # from pylbmisc.dm import Check, Column
    # import pylbmisc as lb
    #
    # df = pd.DataFrame({
    #     "idx":  [100, 101, 102, 103, 103],
    #     "a":    [1, 2, 4, pd.NA, 7],
    #     "b":    [6, 2, 4, pd.NA, 2],
    #     "name": ["asdomar", "asd", "foo", "bar", "baz"]
    # }).convert_dtypes()
    # df["categ"] = lb.dm.to_categorical([1, 2, 3, 1, 2],
    #                                    levels=[1, 2, 3],
    #                                    labels=["A", "B", "C"])
    #
    # ensure = [
    #     Column("idx").is_unique(),
    #     Column("a").lt(3),
    #     Column("a").between(2, 5, include_min=False),
    #     Column("b").isin(range(2, 6)),
    #     Column("categ").isin(["A", "B"]),
    #     Column("categ").notin(["A", "B"]),
    #     Column("name").str_startswith(pat="asd"),
    #     Column("name").str_endswith(pat="ar"),
    #     # cross checks
    #     Check(lambda df: df.a < df.b, "a is not < than b", ["a", "b"], "a_lt_b")
    # ]
    #
    # fails = lb.dm.Validator(ensure=ensure).validate(df, "idx")
    # print(fails)
    """
    def __init__(self, ensure):
        self._ensure = ensure

    def _colname_colval(self, x):
        # breakpoint()
        res = {}
        for colname, value in x.items():
            res[colname] = colname + "=" + _np.where(value.isna(), "NA", value.astype(str))
        return _pd.DataFrame(res).apply(lambda x: ", ".join(x), axis=1)

    def validate(self, df, idcols, return_na_checks=False):
        """Return the failed checks"""
        # breakpoint()
        # normalize if user insert just a single
        if isinstance(idcols, str):
            idcols = [idcols]
        fails = []
        nas = []
        for check in self._ensure:
            success = check.fn(df).convert_dtypes(dtype_backend=_default_dtype_backend)
            # here we need to be sure success is a pd.Series of boolean
            check_ok = isinstance(success, _pd.Series) and is_bool(success)
            if not check_ok:
                msg = f"check with check_id {check.check_id} does" \
                    "not return a pd.Series of bool:" \
                    f"type: {type(success)}, dtype: {success.dtype}"
                raise TypeError(msg)
            na = success.isna()
            fail = ~ success
            # if do display on error or missingness
            df["vars_values"] = \
                self._colname_colval(df[check.display_vars]) if check.display_vars \
                else [""] * df.shape[0]
            sel_cols = idcols + ["vars_values"] if check.display_vars else idcols
            if fail.any():
                fail_df = df.loc[fail, sel_cols]
                fail_df["checks_failures"] = check.errmsg
                fail_df["check_id"] = check.check_id
                fails += [fail_df]
            if na.any():
                nas_df = df.loc[na, sel_cols]
                nas_df["checks_nas"] = check.errmsg
                nas_df["check_id"] = check.check_id
                nas += [nas_df]
        # returning
        fails = _pd.concat(fails)
        nas = _pd.concat(nas)
        col_fails = idcols + ["check_id", "checks_failures", "vars_values"]
        col_nas = idcols + ["check_id", "checks_nas", "vars_values"]
        ret_fail = fails[col_fails].reset_index(drop=True)
        ret_nas = nas[col_nas].reset_index(drop=True)
        rval = (ret_fail, ret_nas) if return_na_checks else ret_fail
        return rval


class Column:
    """Class to create univariate checks on single column. See pa.Check for
    ideas."""
    def __init__(self, name):
        self.name = name

    def gt(self, value):
        return Check(
            fn=lambda df: df.loc[:, self.name] > value,
            errmsg=f"{self.name} is not > {value}",
            display_vars=[self.name],
            check_id=f"{self.name}_gt_{value}"
        )

    def geq(self, value):
        return Check(
            fn=lambda df: df.loc[:, self.name] >= value,
            errmsg=f"{self.name} is not >= {value}",
            display_vars=[self.name],
            check_id=f"{self.name}_geq_{value}"
        )

    def lt(self, value):
        return Check(
            fn=lambda df: df.loc[:, self.name] < value,
            errmsg=f"{self.name} is not < {value}",
            display_vars=[self.name],
            check_id=f"{self.name}_lt_{value}"
        )

    def leq(self, value):
        return Check(
            fn=lambda df: df.loc[:, self.name] <= value,
            errmsg=f"{self.name} is not <= {value}",
            display_vars=[self.name],
            check_id=f"{self.name}_leq_{value}"
        )

    def between(self, min_value, max_value, include_min=True, include_max=True):
        is_above_min = (lambda x: x >= min_value) if include_min \
            else (lambda x: x > min_value)
        is_below_max = (lambda x: x <= max_value) if include_max \
            else (lambda x: x < max_value)
        is_in_between = (lambda df: \
                         (is_above_min(df.loc[:, self.name])) & \
                         (is_below_max(df.loc[:, self.name])))
        range_str = ("[" if include_min else "(") + \
            str(min_value) + ", " + str(max_value) + \
            ("]" if include_max else ")")
        return Check(
            fn=is_in_between,
            errmsg=f"{self.name} is not in {range_str}",
            display_vars=[self.name],
            check_id=f"{self.name}_in_{range_str}"
        )

    def is_unique(self):
        return Check(
            fn=lambda df: ~((df.loc[:, self.name]).duplicated()),
            errmsg=f"{self.name} has duplicated elements",
            display_vars=[self.name],
            check_id=f"{self.name}_is_unique"
        )

    def eq(self, value):
        return Check(
            fn=lambda df: df.loc[:, self.name] == value,
            errmsg=f"{self.name} is not equal to {value}",
            display_vars=[self.name],
            check_id=f"{self.name}_eq_{value}"
        )

    def neq(self, value):
        return Check(
            fn=lambda df: df.loc[:, self.name] != value,
            errmsg=f"{self.name} is equal to {value}",
            display_vars=[self.name],
            check_id=f"{self.name}_neq_{value}"
        )

    def isin(self, values):
        values = set(values)
        return Check(
            fn=lambda df: (df.loc[:, self.name]).isin(values),
            errmsg=f"{self.name} not in {values}",
            display_vars=[self.name],
            check_id=f"{self.name}_isin_admitted_values"
        )

    def notin(self, values):
        return Check(
            fn=lambda df: ~(df.loc[:, self.name].isin(values)),
            errmsg=f"{self.name} belongs to {values}",
            display_vars=[self.name],
            check_id=f"{self.name}_notin_wrong_values"
        )

    def str_startswith(self, **kwargs):
        return Check(
            fn=lambda df: (df.loc[:, self.name]).str.startswith(**kwargs),
            errmsg=f"{self.name} does not start with {kwargs['pat']}",
            display_vars=[self.name],
            check_id=f"{self.name}_startswith_{kwargs['pat']}"
        )

    def str_endswith(self, **kwargs):
        return Check(
            fn=lambda df: (df.loc[:, self.name]).str.endswith(**kwargs),
            errmsg=f"{self.name} does not end with {kwargs['pat']}",
            display_vars=[self.name],
            check_id=f"{self.name}_endswith_{kwargs['pat']}"
        )


    def str_contains(self, **kwargs):
        return Check(
            fn=lambda df: (df.loc[:, self.name]).str.contains(**kwargs),
            errmsg=f"{self.name} does not contain {kwargs['pat']}",
            display_vars=[self.name],
            check_id=f"{self.name}_contains_{kwargs['pat']}"
        )


    def str_match(self, **kwargs):
        return Check(
            fn=lambda df: (df.loc[:, self.name]).str.match(**kwargs),
            errmsg=f"{self.name} does not match {kwargs['pat']}",
            display_vars=[self.name],
            check_id=f"{self.name}_matches_{kwargs['pat']}"
        )



if __name__ == "__main__":
    import doctest
    doctest.testmod()



import pandas as _pd


def _index(x):
    return [ind - 1 for ind in x]


def eortc_qlqc30(x):
    """Scoring for EORTC QLQ-C30

    Calculate the three scores: qol (higher better), functional scale (higher
    better), symptoms scale (lower better),
    following scoring manual available here
    https://www.eortc.org/app/uploads/sites/2/2018/02/SCmanual.pdf

    Parameters
    ----------
    x: pd.DataFrame
        data frame of 30 QLQ-C30 items, one item per column (patient in row)

    Returns
    -------
    rval: pd.DataFrame
        data frame with 3 columns, qol (general QOL), funct and sympts
    """
    qdv_items = [29, 30]
    funct_items = list(range(1, 7+1)) + list(range(20, 27+1))
    sympts_items = list(range(8, 19+1)) + [28]
    # extract data
    qdv = x.iloc[:, _index(qdv_items)]
    funct = x.iloc[:, _index(funct_items)]
    sympts = x.iloc[:, _index(sympts_items)]
    funct_score = (1 - ((funct.mean(axis=1) - 1)/3)) * 100
    qdv_score = ((qdv.mean(axis=1) - 1)/6) * 100
    sympts_score = ((sympts.mean(axis=1) - 1)/3) * 100
    res = _pd.concat([qdv_score, funct_score, sympts_score], axis=1)
    return res.rename({0: "qol", 1: "funct", 2: "sympts"}, axis=1)



"""
Datasets related utilities (list, load).
"""

import pandas as _pd
from importlib import resources as _resources
from pylbmisc.dm import _default_dtype_backend

_dataset_dir = _resources.files("pylbmisc") / "datasets"


def ls() -> list[str]:
    """List available datasets.

    Examples
    --------
    >>> import pylbmisc as lb
    >>> lb.datasets.ls()
    """
    files = sorted(_dataset_dir.rglob("*.csv"))
    fnames = [str(f.name).replace(".csv", "") for f in files]
    return fnames


def load(dname: str = "beetles1", **kwargs) -> _pd.DataFrame:
    """Load an available dataset.

    Parameters
    ----------
    fname: str
        string coming from lb.datasets.ls()
    kwargs: Any
        named paramers passed to pd.read_csv

    Examples
    --------
    >>> import pylbmisc as lb
    >>> lb.datasets.ls()
    >>> df = lb.datasets.load("laureati.csv")
    """
    data_file = _dataset_dir / (dname + ".csv")
    doc_file = _dataset_dir / (dname + ".txt")
    data = _pd.read_csv(data_file,
                        engine="python",
                        dtype_backend=_default_dtype_backend,
                        **kwargs)
    with open(doc_file, "r") as f:
        doc = f.read()
    data.__doc__ = doc
    return data



"""
surv
========================================
Survival analysis function and utilities.
"""

import numpy as _np
import pandas as _pd
import matplotlib.pyplot as _plt

from lifelines import KaplanMeierFitter as _KaplanMeierFitter
from lifelines import CoxPHFitter as _CoxPHFitter
from lifelines.plotting import add_at_risk_counts as _add_at_risk_counts
from lifelines.statistics import multivariate_logrank_test \
    as _multivariate_logrank_test
from lifelines.utils import qth_survival_times as _qth_survival_times
from pylbmisc.dm import to_integer as _to_integer
from pylbmisc.dm import is_datetime as _is_datetime
from pylbmisc.stats import p_format as _p_format
from warnings import warn as _warn


def _estquant(fit, quantiles):
    """Return estimates and quantiles of survival function with confidence
    intervals"""
    estimates = _pd.concat([fit.survival_function_, fit.confidence_interval_],
                           axis="columns")
    quant = _qth_survival_times(quantiles, estimates)
    quant = quant.reset_index()
    quant.columns = ["Quantile", "Estimate", "Lower", "Upper"]
    estimates = estimates.reset_index()  # take time as variable
    estimates.columns = ["time", "Estimate", "Lower", "Upper"]
    return estimates, quant


def km(time,
       status,
       group=None,
       plot=True,
       plot_censored=True,
       plot_censored_style={"ms": 5,  "marker": "|"},
       plot_legend_loc=None, # TODO IMPLEMENT
       plot_logrank=True,
       ylim=(-0.05, 1.05),
       ylab="Survival probability",
       xlab="Time",
       counts=["Events"],
       xticks=None,
       ci_alpha=0.3,
       quantiles=[0.5]):
    """Kaplan-Meier estimates, logrank test and Hazard ratios.

    Does the Kaplan-Meier plot with logrank

    Parameters
    ----------
    time: time
        the time
    status: dichotomic
        the event indicator
    group: categorical variable
        a grouping variable used to create different survival groups/function
    plot: logical
        wether to do the plot or not (just return estimates)
    plot_censored: logical
        plot ticks of censored observations
    plot_censored_style: dict
        dict of configs given to matplotlib.lines.Line2D
    plot_legend_loc: None, str or pair of floats
        positioning of legend: can None (use defaults: no plotting for one
        curve, "best" for 2+), "none": dont plot, any loc parameter for
        ax.legend (eg "lower center"), "outside": bbox_to_anchor to the right
        by default upper-right if two or more groups are
        available, none if only one group is plotted
    plot_logrank: bool
         add logrank to the plot
    ylim: tuple
        as in R
    ylab: ylab
        as in R
    xlab: xlab
        as in R
    counts: list
         containing ["At risk", "Events", "Censored"]
    xticks: a slice
         slice used for plotting (loc) eg slice(5) plots up to time = 5
    ci_alpha: percentage
         shading (alpha) for confidence interval  (set to 0 for no CI)
    quantiles: list[float]
         list of quantiles of survival function to be returned (def: median)

    Returns
    -------
    dict
        dict with some results

    """
    if plot:
        fig, ax = _plt.subplots()
        plot_at_risk = bool(counts)
    if group is None:  # --------------------single curve --------------------
        kmf = _KaplanMeierFitter()
        df = _pd.DataFrame({"time": time, "status": status}).dropna()  # handle
        fit = kmf.fit(df["time"], df["status"])
        estimates, quants = _estquant(fit, quantiles)
        if plot:
            # https://lifelines.readthedocs.io/en/latest/fitters/univariate/KaplanMeierFitter.html
            ax = fit.plot_survival_function(
                loc=xticks,
                show_censors=plot_censored,
                censor_styles=plot_censored_style,
                ci_alpha=ci_alpha)
            ax.set_ylim(ylim)
            ax.set_ylabel(ylab)
            ax.set_xlabel(xlab)
            # legend: avoid by default for 1 group
            if (plot_legend_loc is None or plot_legend_loc == "none"):
                lgnd = ax.legend()
                lgnd.set_visible(False)
            elif (plot_legend_loc == "outside"):
                lgnd = ax.legend(bbox_to_anchor=(1, 1), loc="upper left")
            else:
                lgnd = ax.legend(loc=plot_legend_loc)
            # number at risk
            if plot_at_risk:
                _add_at_risk_counts(fit, labels=["All"],
                                    rows_to_show=counts,
                                    ax=ax, fig=fig)
                _plt.tight_layout()
            fig.show()
        return {
            "fit": fit,
            "estimates": estimates,
            "quantiles": quants
        }
    else:  # ---------------------- several curves ----------------------------
        try:
            categs = group.cat.categories.to_list()
        except AttributeError:
            msg = f"Group must be a pandas categorical variable: now {group.dtype}"
            raise AttributeError(msg)

        if len(categs) < 2:
            msg = "Group must have at least two categories."
            raise Exception(msg)

        # do fit for all categories
        fits = {}               # fits
        estimates = {}          # survival estimates
        quants = []             # quantiles

        df = _pd.DataFrame({
            "time": time,
            "status": status,
            "group": group}).dropna()
        df["group"] = df.group.cat.remove_unused_categories()
        new_categs = df.group.cat.categories.to_list()

        if len(new_categs) < len(categs):
            removed_categ = set(categs) - set(new_categs)
            removed_categ_str = ", ".join(removed_categ)
            msg = "Some categories were removed due" \
                f" to missingness: {removed_categ_str}."
            _warn(msg)

        # Kaplan-meier fit and quantiles
        for categ in new_categs:
            kmf = _KaplanMeierFitter(label=ylab)
            mask = df["group"] == categ
            fits[categ] = f = kmf.fit(df.loc[mask, "time"],
                                      df.loc[mask, "status"],
                                      label=categ)
            e, q = _estquant(f, quantiles)
            estimates[categ] = e
            q.insert(0, "Group", categ)
            quants.append(q)
            if plot:
                kmf.plot_survival_function(
                    ax=ax,
                    show_censors=plot_censored,
                    censor_styles=plot_censored_style,
                    loc=xticks,
                    ci_alpha=ci_alpha)
        quants = _pd.concat(quants)
        # lograng test
        lr = _multivariate_logrank_test(df["time"], df["group"], df["status"])
        # Cox HR:
        # need to construct dummies https://stackoverflow.com/questions/60114847
        dummies = _pd.get_dummies(df.group, drop_first=True)
        df_cox = _pd.concat([df[["time", "status"]], dummies], axis=1)
        cph = _CoxPHFitter().fit(df_cox, 'time', 'status')
        cox_keep = ['coef', 'coef lower 95%', 'coef upper 95%', 'p']
        cox_res = cph.summary[cox_keep]
        cox_res.iloc[:, 0:3] = cox_res.iloc[:, 0:3].apply(_np.exp)
        cox_res.columns = ["HR", "lower.95", "upper.95", "p"]
        cox_res.index = cox_res.index.set_names("group")
        # aggiungo numerosità e gruppo di base per forest plot
        freqs = _pd.DataFrame({"n": df.group.value_counts(dropna=True, sort=False)})
        freqs.index = freqs.index.set_names("group")
        cox_res = _pd.merge(freqs, cox_res,
                            left_index=True, right_index=True,
                            how="left")
        cox_res.iloc[0, 1] = 1 # HR for base group
        
        # plotting
        if plot:
            ax.set_ylim(ylim)
            ax.set_ylabel(ylab)
            ax.set_xlabel(xlab)

            # legend: by default best for more than 1 group
            if (plot_legend_loc is None):
                lgnd = ax.legend(loc="best")
            elif ("none" == plot_legend_loc):
                lgnd = ax.legend()
                lgnd.set_visible(False)
            elif ("outside" == plot_legend_loc):
                lgnd = ax.legend(bbox_to_anchor=(1, 1), loc="upper left")
            else:
                lgnd = ax.legend(loc=plot_legend_loc)
            # number at ris
            if plot_at_risk:
                _add_at_risk_counts(*list(fits.values()),
                                    rows_to_show=counts,
                                    ax=ax, fig=fig)
                _plt.tight_layout()

            if plot_logrank:
                lr_string = (f"logr: {lr.test_statistic:.3f}, "
                             f"df: {lr.degrees_of_freedom}, "
                             f"p: {_p_format(lr.p_value)}")
                ax.set_title(lr_string)
            fig.show()

        return {
            "fit": fits,
            "estimates": estimates,
            "quantiles": quants,
            "logrank": lr,
            "hr": cox_res
        }


# --------------------------------------------------------------------------------------


def _date_check(x: _pd.Series, name: str = "date", can_be_none: bool = False):
    if (x is None) and (not can_be_none):
        msg = f"{name} can't be None"
        raise Exception(msg)
    if not _is_datetime(x):
        msg = f"{name} must be an allowed type (datetime)"
        raise Exception(msg)


def _check_sequential_dates(x: _pd.DataFrame):
    int_dates = x.apply(_to_integer, axis=1)
    na_dates = x.isna()
    int_dates[na_dates] = _pd.NA
    not_sequential = (int_dates.diff(axis=1) < 0)
    not_sequential_mask = not_sequential.any(axis=1)
    if not_sequential_mask.any():
        msg = "Some dates are not sequential."
        _warn(msg)
        with _pd.option_context("display.max_rows", None,
                                "display.max_columns", None):
            print(_pd.concat([x, not_sequential],
                             axis=1).loc[not_sequential_mask, :])


def _check_negative_times(s, t, outcome, verbose=True):
    status = s.copy()
    time = t.copy()
    neg_times = time < 0
    if neg_times.any():
        status[neg_times] = _pd.NA
        time[neg_times] = _pd.NA
        if verbose:
            msg = f"Some {outcome} times are < 0. Setting to NA" \
                " both time and status below. Please fix the dates."
            _warn(msg)
            with _pd.option_context("display.max_rows", None,
                                    "display.max_columns", None):
                print(_pd.DataFrame({"status": status,
                                     "time": time}).loc[neg_times, :])
    return status, time


def tteep(start_date=None,
          prog_date=None,
          death_date=None,
          last_fup=None,
          ep=["os", "pfs", "ttp"],
          verbose=True
          ):
    """This function calculates common oncology time to event end-points
    (Overall Survival, Progression Free Survival, Time to Relapse).

    Overall survival (OS) is computed as time from start_date to
    death_date. Time of patients who did not experienced the event (with
    missing death_date) is censored at the time of last follow up
    (last_fup).

    Progression free survival (PFS) is calculated as time from start_date
    to prog_date or death_date, whichever comes first. Time of
    patients who did not experienced any event (missing both prog_date and
    death_date) is censored at the time of last follow up (last_fup).

    Time to progression (TTP) is computed as time from start_date to
    prog_date. Time of patients who did not experienced the event (missing
    prog_date) is censored at the time of last follow up (last_fup) or death
    (death_date) whichever comes first.

    Parameters
    ----------
    start_date: Date
        starting follow up date
    prog_date: Date
        progression date
    death_date: Date
        death date
    last_fup: Date
        last follow up date
    ep:
        which end points to calculate, default to ["os","pfs","ttp"]

    Examples
    --------
    >>> import pandas as pd
    >>> from pylbmisc.dm import to_date
    >>> df = pd.DataFrame({
    ...    "start_date" : to_date(["1900-01-01", "1900-01-01", "1900-01-01",
    ...                            "1900-01-01", pd.NA, "1900-01-01",
    ...                            "1900-04-01", pd.NA]),
    ...    "prog_date"  : to_date(["1900-03-01", "1900-03-01", pd.NA,
    ...                            pd.NA, pd.NA, pd.NA, "1900-02-01", pd.NA]),
    ...    "death_date" : to_date(["1900-06-01",  pd.NA, "1900-06-01",
    ...                            pd.NA, pd.NA, pd.NA, "1900-06-01", pd.NA]),
    ...    "last_fup"   : to_date(["1900-06-01", "1900-12-31", "1900-06-01",
    ...                            "1900-12-31", "1900-12-31", pd.NA,
    ...                            "1900-06-01", pd.NA])
    ... })
    >>> print(df)
      start_date  prog_date death_date   last_fup
    0 1900-01-01 1900-03-01 1900-06-01 1900-06-01
    1 1900-01-01 1900-03-01        NaT 1900-12-31
    2 1900-01-01        NaT 1900-06-01 1900-06-01
    3 1900-01-01        NaT        NaT 1900-12-31
    4        NaT        NaT        NaT 1900-12-31
    5 1900-01-01        NaT        NaT        NaT
    6 1900-04-01 1900-02-01 1900-06-01 1900-06-01
    7        NaT        NaT        NaT        NaT
    >>> tteep(df.start_date, df.prog_date, df.death_date, df.last_fup,
    ...       verbose=False)
       os_status  os_time  pfs_status  pfs_time  ttp_status  ttp_time
    0          1    151.0           1      59.0           1      59.0
    1          0    364.0           1      59.0           1      59.0
    2          1    151.0           1     151.0           0     151.0
    3          0    364.0           0     364.0           0     364.0
    4       <NA>      NaN        <NA>       NaN        <NA>       NaN
    5       <NA>      NaN        <NA>       NaN        <NA>       NaN
    6          1     61.0        <NA>       NaN        <NA>       NaN
    7       <NA>      NaN        <NA>       NaN        <NA>       NaN

    Reference
    ----------
    Guidance for Industry, Clinical Trial Endpoints for the approval of cancer
    drugs and biologics, FDA, May 2007

    """

    _date_check(x=start_date, name="start_date", can_be_none=False)
    _date_check(x=last_fup, name="last_fup", can_be_none=False)
    _date_check(x=prog_date, name="prog_date", can_be_none=True)
    _date_check(x=death_date, name="death_date", can_be_none=True)

    n = len(start_date)

    # Create indicator variables based on dates
    prog = (~prog_date.isna()) if (prog_date is not None) \
        else _pd.Series([_pd.NA] * n)
    death = (~death_date.isna()) if (death_date is not None) \
        else _pd.Series([_pd.NA] * n)

    # dates dataframe for checks and other
    all_dates = _pd.DataFrame({
        "start_date": start_date,
        "prog_date": prog_date,
        "death_date": death_date,
        "last_fup": last_fup
    })

    # Check sequential dates
    if verbose:
        _check_sequential_dates(all_dates)

    rval = {}
    if "os" in ep:
        os_status = _to_integer(death)
        os_last_date = _np.where(os_status, death_date, last_fup)
        os_time = (os_last_date - start_date).dt.days
        os_status[os_time.isna()] = _pd.NA
        s, t = _check_negative_times(os_status, os_time,
                                     "OS", verbose=verbose)
        rval["os_status"] = s
        rval["os_time"] = t

    if "pfs" in ep:
        pfs_status = _to_integer(death | prog)
        min_prog_death = _pd.concat([death_date, prog_date], axis=1).min(axis=1)
        pfs_last_date = _np.where(pfs_status, min_prog_death, last_fup)
        pfs_time = (pfs_last_date - start_date).dt.days
        pfs_status[pfs_time.isna()] = _pd.NA
        s, t = _check_negative_times(pfs_status, pfs_time,
                                     "PFS", verbose=verbose)
        rval["pfs_status"] = s
        rval["pfs_time"] = t

    if "ttp" in ep:
        ttp_status = _to_integer(prog)
        min_death_lfup = _pd.concat([death_date, last_fup], axis=1).min(axis=1)
        ttp_last_date = _np.where(ttp_status, prog_date, min_death_lfup)
        ttp_time = (ttp_last_date - start_date).dt.days
        ttp_status[ttp_time.isna()] = _pd.NA
        s, t = _check_negative_times(ttp_status, ttp_time,
                                     "TTP", verbose=verbose)
        rval["ttp_status"] = s
        rval["ttp_time"] = t

    return _pd.DataFrame(rval)


def coxph(df, time, status, formula, **kwargs):
    """ Fit a Cox proportional hazard model.

    Fit a Cox proportional hazard model, with check for proportionality
    assumption using Schoenfeld's residuals.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with variables
    time: chr
         variable name for time in df
    status: chr
         variable name for status in df
    formula: chr
         formula as per lifelines' CoxPHFitter
    kwargs: dict
         other parameters passed to CoxPHFitter.fit

    Returns
    -------
    mod: pd.DataFrame
         DataFrame with the estimated model's HR, ci and pretty printed p-values.

    """
    cph = _CoxPHFitter()
    cph.fit(df=df,
            duration_col=time,
            event_col=status,
            formula=formula,
            **kwargs)
    cols_kept = ['exp(coef)', 'exp(coef) lower 95%', 'exp(coef) upper 95%', 'p']
    mod = cph.summary[cols_kept]
    mod.columns = ["HR", "Low 95%CI", "Up 95%CI", "p"]
    mod["p"] = _p_format(mod["p"])
    # schoenfeld ph test
    cph.check_assumptions(df,
                          advice=False,
                          # p_value_threshold=0.05, # by default at 0.01
                          show_plots=False)
    # returning
    return mod


def median_fup(time, status, group=None):
    r"""
    Calculate median follow up using reverse Kaplan-Meier method

    Parameters
    ----------
    time: pd.Series only?
        the time variable
    status: pd.Series only?
        the status variable (eg os_status)
    group: None or categorical
        groups to calculate stratified medians

    References
    ----------
    Schemper, Michael, and Terry L. Smith. “A Note on Quantifying Follow-up in
    Studies of Failure Time.” Controlled Clinica Trials 17, no. 4 (1996):
    343–46. https://doi.org/10.1016/0197-2456(96)00075-X

    Examples
    --------
    >>> import pylbmisc as lb
    >>> ov = lb.datasets.load("ovarian.csv")
    >>> median_fup(time=ov.survtime/365.25, status=ov.surv)
       Quantile  Estimate     Lower      Upper
    0       0.5  9.604381  9.029432  10.110883
    >>> ov["group"] = lb.dm.to_categorical(ov.histo_cl)
    >>> median_fup(time=ov.survtime/365.25, status=ov.surv, group=ov.group)
            Group  Quantile  Estimate     Lower      Upper
    0      serous       0.5  9.637235  9.029432  10.234086
    0  not serous       0.5  9.366188  8.295688  10.390144
    """
    km_res = km(
        time=time,
        status=1-status,
        group=group,
        plot=False,
        quantiles=[0.5]
    )
    return km_res["quantiles"]


def censor_at(time, status, censoring_time, name_prefix=""):
    r"""Censor observation at a given time

    Parameters
    ----------
    time: pd.Series
        time variable
    status: pd.Series
        status variable
    censoring_time: int
        integer where to censor
    name_prefix: char
        string to prefixto the outputted dataset (eg "os_")

    Examples
    --------
    >>> import pylbmisc as lb
    >>> ov = lb.datasets.load("ovarian.csv")
    >>> pd.concat([
    >>>    ov[["survtime","surv"]],
    >>>    censor_at(time=ov.survtime, status=ov.surv, censoring_time=500)
    >>> ], axis=1)
    >>>
    >>> time = pd.Series([100, 150, 200])
    >>> timena = pd.Series([100, 150, np.nan])
    >>> status = pd.Series([1, 0, 0])
    >>> statusna = pd.Series([0, 1, np.nan])
    >>> censor_at(time, statusna, 120)
    >>> censor_at(timena, status, 120)
    """
    to_censor = time > censoring_time
    time_censored = _np.where(to_censor, censoring_time, time)
    status_censored = _np.where(
        to_censor,
        _np.where(_pd.isna(status), _pd.NA, 0),
        status)
    time_varname = f"{name_prefix}time_cens{censoring_time}"
    status_varname = f"{name_prefix}status_cens{censoring_time}"
    rval = _pd.DataFrame({
        time_varname: time_censored,
        status_varname: _to_integer(status_censored)
    })
    return rval


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import re


def main():
    with open("doc_export.Rout", "r") as f:
        doc = f.read()

    doc_splitted = doc.split("\n$")[1:]
    name_re = re.compile("^(.+)\n")
    # filename = name_re.match(doc_splitted[1]).group(0).replace("\n","")  + ".txt"
    # # name_re.sub(r"\1", doc_splitted[1])
    # actual_doc = name_re.sub(r"", doc_splitted[1])

    for man in doc_splitted:
        fname = name_re.match(man).group(0).replace("\n","")  + ".txt"
        actual_doc = name_re.sub(r"", man)
        with open(fname, "w") as f:
            print(actual_doc, file=f)


if __name__ == "__main__":
    main()
